from skopt import Optimizer

import ray
from ray.tune.suggest.skopt import SkOptSearch

from hpogrid.search_space.skopt_space import SkOptSpace



class SkOptAlgoWrapper():
    def __init__(self):
        self.algorithm = None
    def create(self, metric, mode, search_space, **args):
        search_space = SkOptSpace(search_space).search_space
        optimizer = Optimizer(search_space)
        labels = [hp.name for hp in search_space]
        self.algorithm = SkOptSearch(optimizer, labels, metric=metric, mode=mode, **args)
        return self.algorithm


import json
import os

from hpogrid.components.defaults import kConfigFormat, kProjConfigFormat, kHPOGridMetadataFormat


def type2str(typevar):
    if isinstance(typevar, tuple):
        return ', '.join([t.__name__ for t in typevar])
    else:
        return typevar.__name__
    
    
def validate(config, config_format):
    for key in config_format:
        if key in config:
            # check if the value type of the config is correct
            value_type = config_format[key]['type']
            if not isinstance(config[key], value_type):
                raise ValueError('The value of "{}" must be of type {}'.format(key, type2str(value_type)))
            # check if the value of the config is allowed
            if ('choice' in config_format[key]) and (config[key] not in config_format[key]['choice']):
                raise ValueError('The value of "{}" must be one of the followings: {}'.format(
                                 key, str(config_format[key]['choice']).strip('[]')))
        else:
            if config_format[key]['required']:
                raise ValueError('The required item "{}" is missing from the configuration'.format(key))
            # fill in default config if not specified
            if 'default' in config_format[key]:
                print('Info: Added the item "{}" with default value {} to the configuration'.format(
                    key, str(config_format[key]['default'])))
                config[key] = config_format[key]['default']
    for key in config:
        if key not in config_format:
            raise ValueError('Unknown item "{}" found in the configuration'.format(key)) 
    return config

def validate_config(config, config_type):
    if config_type == 'search_space':
        return validate_search_space(config)
    if config_type not in kConfigFormat:
        raise ValueError('Unknown configuration type "{}"'.format(config_type))
    config_format = kConfigFormat[config_type]
    print('Info: Validating {} configuration'.format(config_type))
    validate(config, config_format)
    print('Info: Successfully validated {} configuration'.format(config_type))
    return config 

def validate_project_config(config):
    print('INFO: Checking validity of configurations')
    config = validate(config, kProjConfigFormat)
    for sub_config in ['model_config', 'search_space', 'hpo_config', 'grid_config']:
        config_type = sub_config.replace('_config', '')
        config[sub_config] = validate_config(config[sub_config], config_type)
    return config
        
def validate_search_space(search_space):
    print('Info: Validating search space configuration...')
    config_format = kConfigFormat['search_space']
    sampling_methods = config_format.keys()
    kHPKeys = set(['method', 'dimension'])
    for hp in search_space:
        if (not isinstance(search_space[hp], dict)) or (set(search_space[hp].keys()) != kHPKeys):
            raise ValueError('Each hyperparameter must be defined by a dictionary containing the keys "method" and "dimension"')
        if not isinstance(search_space[hp]['dimension'], dict):
            raise ValueError('Dimension of a hyperparameter must be a dictionary containing the parameters of its sampling method')
        method = search_space[hp]['method']
        # check if hyperparameter sampling method is supported
        if method not in sampling_methods:
            raise ValueError('Sampling method "{}" is not supported. '
                             'Supported methods are: {}'.format(method,','.join(sampling_methods)))
        for arg in config_format[method]:
            if arg in search_space[hp]['dimension']:
                # check if the argument type of the sampling method is correct
                value_type = config_format[method][arg]['type']
                if not isinstance(search_space[hp]['dimension'][arg], value_type):
                    raise ValueError('The value of argument "{}" for the sampling method '
                          '"{}" (for hyperparameter "{}") must be of type {}'.format(
                        arg, method, hp, type2str(value_type)))
            else:
                # check if a required argument for the sampling method is missing
                if (config_format[method][arg]['required']):
                    raise ValueError('Missing argument "{}" for the sampling method "{}" '
                                     'for the hyperparameter "{}"'.format(arg, method, hp))
                # set the argument to its default value if not specified
                if ('default' in config_format[method][arg]):
                    print('Info: The argument "{}" for the sampling method "{}" for'
                          ' the hyperparameter "{}" will be set to its default value {}'.format(
                          arg, method, hp, config_format[method][arg]['default']))
                    search_space[hp]['dimension'][arg] = config_format[method][arg]['default']
    print('Info: Successfully validated search space configuration')
    return search_space


def validate_job_metadata(data):
    if not isinstance(data, dict):
        return False
    return all(key in kHPOGridMetadataFormat for key in data)
    
def parse_configuration(config):
    validate_project_config(config)
    


import ray
from ray.tune.suggest.bohb import TuneBOHB

from hpogrid.search_space.bohb_space import BOHBSpace


class BOHBAlgoWrapper():
    def __init__(self):
        self.algorithm = None
    def create(self, metric, mode, search_space, **args):
        search_space = BOHBSpace(search_space).search_space
        self.algorithm = TuneBOHB(search_space, metric=metric, mode=mode, **args)
        return self.algorithm



import ray
from ray.tune.suggest.hyperopt import HyperOptSearch

from hpogrid.search_space.hyperopt_space import HyperOptSpace

class HyperOptAlgoWrapper():
    def __init__(self):
        self.algorithm = None
    def create(self, metric, mode, search_space, **args):
        search_space = HyperOptSpace(search_space).search_space
        self.algorithm = HyperOptSearch(
            search_space,
            metric=metric,
            mode=mode, **args)
        return self.algorithm


import sys, os
import ssl
import json
#import re
import fnmatch

from tqdm import tqdm
import argparse
import datetime
import pandas as pd
from pdb import set_trace

try:
    from urllib.parse import urlencode
    from urllib.parse import urljoin
    from urllib.request import urlopen, Request
    from urllib.error import HTTPError, URLError
except ImportError:
    from urllib import urlencode
    from urllib2 import urlopen, Request, HTTPError, URLError
    from urlparse import urljoin

from tabulate import tabulate
from colorama import Fore, Back, Style

try:
    from pandatools import PBookCore
    from pandatools import Client
    from pandatools.PBookCore import PsubUtils
except:
    raise ImportError('No module named \'pandatools\'. Please do '
        '\'setupATLAS\' and \'lsetup panda\' first ')


kPandaTaskManagerMethods = ['show', 'kill', 'retry', 'killAndRetry']

kPanDA_URL = 'https://bigpanda.cern.ch/'
kPanDA_TaskURL = 'https://bigpanda.cern.ch/tasks/'
kPanDA_JobURL = 'https://bigpanda.cern.ch/jobs/'

kPanDAVerifyPath = {
    'cafile' : '/etc/pki/tls/cert.pem',
    'capath' : '/etc/pki/tls/certs',
}

kHeaders = {'Accept': 'application/json', 'Content-Type':'application/json'}


kTaskActiveSuperstatusList = ['running', 'submitting', 'registered', 'ready']
KTaskFinalSuperstatusList = ['finished', 'failed', 'done', 'broken', 'aborted']
kStandardColumns = ['jeditaskid', 'status', 'pctfinished','taskname']
kOutputColumns = ['jeditaskid', 'status', 'taskname', 'computingsite', 'site', 'metastruct']

kStatusLen = 8
kLine = '_'*90
kStrf = '{jeditaskid:>10}  {status:<{l}}  {pctfinished:>5} {taskname}'
kHead = kStrf.format(l=kStatusLen, status='Status', jeditaskid='JediTaskID',
    pctfinished='Fin%', taskname='TaskName') +'\n' + kLine

RED = '\033[0;91m'
GREEN = '\033[92m'
YELLOW = '\033[93m'
BLUE = '\033[94m'
MAGENTA = '\033[95m'
CYAN = '\033[96m'
ENDC = '\033[0m'
BOLD = '\033[1m'
UNDERLINE = '\033[4m'
BLINK = '\033[5m'
kColorDict = {
    'running': BLUE + BOLD,
    'submitting': CYAN,
    'registered': MAGENTA,
    'ready': MAGENTA,
    'pending': MAGENTA,
    'done': GREEN,
    'finished': YELLOW,
    'broken': RED + BOLD ,
    'exhausted': RED,
    'aborted': RED,
    'failed': RED,
    }


class PandaTaskManager():

    def __init__(self, enable_parser=False):
        print('INFO: Verifying Grid Proxy...')
        self.pbook = PBookCore.PBookCore()
        print('INFO: Verification is Successful.')
        self.context = ssl.SSLContext()
        #self.context = ssl._create_unverified_context()
        self.context.load_verify_locations(**kPanDAVerifyPath)

        # call functions via hpogrid executable
        if enable_parser:
            self.run_parser()

    def run_parser(self):
        base_parser = self.get_parser()
        args = base_parser.parse_args(sys.argv[2:3])
        method = args.method
        method_parser = self.get_parser(method)
        if not method_parser:
            print('Unrecognized method: {}'.format(method))
            base_parser.print_help()
            exit(1)
        args = method_parser.parse_args(sys.argv[3:])
        getattr(self, method)(args)

    def get_parser(self, method=None):
        parser = None
        if method == None:
            parser = argparse.ArgumentParser()
            parser.add_argument('method', help='Method to call',
                choices=kPandaTaskManagerMethods)    
        elif method == 'show':
            parser = argparse.ArgumentParser(description='show PanDA task status',
                        formatter_class=argparse.RawDescriptionHelpFormatter)      
            parser.add_argument('-u','--username',
                help='filter tasks by username')
            parser.add_argument('-l','--limit', type=int, default=1000,
                help='the maximum number of tasks to query')
            parser.add_argument('-d', '--days', type=int, default=30,
                help='filter tasks within the recent N days')
            parser.add_argument('-n', '--taskname', 
                help='filter tasks by taskname (accept wildcards)')
            parser.add_argument('-i', '--jeditaskid',
                help='only show the task with the specified jeditaskid')
            parser.add_argument('-m', '--metadata', action='store_true',
                help='print out the metadata of a task')
            parser.add_argument('-s', '--status', nargs='+',
                help='filter tasks by task status')
            parser.add_argument('-y', '--sync', action='store_true',
                help='force no caching on the PanDA server')
            parser.add_argument('-r', '--taskid_range',
                help='filter tasks by jeditaskid range')
            parser.add_argument('-o', '--output',
                help='output result with the filename if specified')
            parser.add_argument('-c', '--outcol',
                help='data columns to be saved in output',
                default=kOutputColumns)
        elif (method in ['kill', 'retry', 'killAndRetry']):
            parser = argparse.ArgumentParser(description='kill/retry a PanDA task',
                        formatter_class=argparse.RawDescriptionHelpFormatter)   
            parser.add_argument('jeditaskid', help='jeditaskid of the task')   

        return parser

    def get_response(self, req, out_json=True):
        response = urlopen(req, context=self.context).read().decode('utf-8')
        if out_json:
            response = json.loads(response)
        return response

    def get_request(self, base_url, params, headers=kHeaders):
        url = '{0}?{1}'.format(base_url, urlencode(params))
        req = Request(url, headers=headers)
        return req

    def get_request_url(self, req):
        url = req.get_full_url()
        return url


    def query_job_metadata(self, taskids):
        metadata = []
        pbar = tqdm(taskids)
        for tid in pbar:
            pbar.set_description("Processing taskid %s" % tid)
            metadata.append(Client.getUserJobMetadata(tid, verbose=0)[1])
        return metadata

    #deprecated
    def query_by_jeditaskid(self, taskids):
        params = {
            'json': 1,
            'status': 'done',
            'extra': 'metastruct'
        }
        if not isinstance(taskids, list):
            taskids = [taskids]
        response = []
        pbar = tqdm(taskids)
        for tid in pbar:
            pbar.set_description("Processing taskid %s" % tid)
            params['jeditaskid'] = tid
            req = self.get_request(kPanDA_TaskURL, params)
            response += self.get_response(req)
        return response



    def query_tasks(self, taskname=None, jeditaskid=None, username=None, limit=1000,
                    status=None, superstatus=None, days=30, sync=False, metadata=False, 
                    taskid_range=None, **args):
        if username is None:
            username = self.pbook.username
        
        _status = status if not isinstance(status, list) else None
        
        params = { 
        'json': 1,
        'limit':limit,
        'jeditaskid': jeditaskid,
        'username': username,
        'taskname': taskname,
        'status': _status,
        'superstatus': superstatus,
        'days': days
        }

        params = {k: v for k, v in params.items() if v is not None}
    
        if sync:
            params['sync'] = int(time.time())
        
        if metadata:
            params['extra'] = 'metastruct'
        

        print('INFO: Showing only max {} tasks in last {} days.'
            ' One can set "--days N" to see tasks in last N days,'
            ' \n      and "--limit M" to see at most M latest tasks'
            .format(limit, days))
        response = None
        req = self.get_request(kPanDA_TaskURL, params)
        response = self.get_response(req)

        n_jobs = len(response)
        
       # set_trace()
        if taskid_range:
            response = self.filter_range(response, taskid_range)
        if isinstance(status, list):
            response = self.filter_status(response, status)
        
        new_n_jobs = len(response)
        if metadata and (n_jobs > 100):
            print('INFO: Fetching metadata for {} jobs '
                'which is greater than the maximum limit (100). '
                'Number of jobs after filtering task id range and'
                ' task status is: {}.'
                'Need to fetch metadata individually which '
                'will be slower...\n'
                'TIPS: Try to filter tasks by their task names to '
                'remove irrelevant tasks. For example, if your '
                'tasks have names of the form user.USERNAME.'
                'PROJECTNAME.SOMETHING.ELSE, then you can use '
                '--taskname *.PROJECTNAME.* to filter jobs'.format(n_jobs,new_n_jobs))
            taskids = [res['jeditaskid'] for res in response]
            job_metadata_list = self.query_job_metadata(taskids)
            for task, job_metadata in zip(response, job_metadata_list):
                task['jobs_metadata'] = job_metadata
            """
            overflow_task_ids = [res['jeditaskid'] for res in response[100:]]
            params['limit'] = 100
            req = self.get_request(kPanDA_TaskURL, params)
            response = self.get_response(req)
            response += self.query_job_metadata(overflow_task_ids)
            """

        return response

    def set_status_color(self, params):
        status = params['status']
        color = kColorDict.get(status,ENDC)
        params['status'] = color+status+ENDC
        nonprlen = len(color) + len(ENDC)
        params['l'] = kStatusLen+nonprlen


    def print_standard(self, datasets):
        filtered_data = []
        print(kHead)
        for task in datasets:
            params = { k:task[k] for k in kStandardColumns if not k.startswith('pct')}
            params_pct = { k:'{0}%'.format(task['dsinfo'][k]) for k in kStandardColumns if k.startswith('pct')}
            params.update(params_pct)
            filtered_data.append(params.copy())
            self.set_status_color(params)
            print(kStrf.format(**params))
    @staticmethod
    def extract_metadata(datasets):
        metadata = {}
        # loop over tasks
        for task in datasets:
            # check if jobs metadata exists in a dataset
            if task.get('jobs_metadata',None):
                task_id = task.get('jeditaskid', None)
                if task_id is None:
                    raise ValueError('Unable to extract jeditaskid from url response')
                task_name = task.get('taskname', None)
                if task_name is None:
                    raise ValueError('Unable to extract taskname from url response')
                metadata[task_id] = {'taskname': task_name, 'jobs_metadata':{} }
                # loop over jobs in a task
                for job_id in task['jobs_metadata']:
                    # fill user job metadata from a job
                    metadata[task_id]['jobs_metadata'][job_id] = task['jobs_metadata'][job_id].get('user_job_metadata', {})
                        
        return metadata
    
    @staticmethod
    def print_metadata(datasets):
        metadata = PandaTaskManager.extract_metadata(datasets)
        for task_id in metadata:
            print('INFO: Printing metadata for task: {} (JediTaskID: {})'.format(metadata[task_id]['taskname'], task_id))
            for job_id in metadata[task_id]['jobs_metadata']:
                print('Job ID: {}'.format(job_id))
                print('User Job Metadata:\n{}\n'.format(metadata[task_id]['jobs_metadata'][job_id]))

    # filter task by jeditaskid range
    @staticmethod
    def filter_range(datasets, frange):
        r = frange.split('-')
        if frange[0] == '-':
            datasets = [ds for ds in datasets if int(ds['jeditaskid']) <= int(r[1])]
        elif frange[-1] == '-':
            datasets = [ds for ds in datasets if int(ds['jeditaskid']) >= int(r[0])]
        else:
            datasets = [ds for ds in datasets if int(ds['jeditaskid']) in range(int(r[0]),int(r[1]))]   
        return datasets
    
    # filter task by status
    @staticmethod
    def filter_status(datasets, status):
        if not isinstance(status, list):
            status = [status]
        datasets = [ds for ds in datasets if ds.get('status', None) in status]
        return datasets    

    def filter_datasets(self, datasets, col=kOutputColumns, out_type=list):
        df = pd.DataFrame(datasets)
        df = df.filter(col).transpose()
        if out_type is list:
            df_dict = df.to_dict()
            return [df_dict[key] for key in df_dict]
        elif out_type is dict:
            return df.to_dict()
        else:
            raise TypeError('output data type {} is not supported'.format(out_type))

    def save_datasets(self, datasets, out_path, col=kOutputColumns):
        with open(out_path, 'w') as outfile:
            out_dataset = self.filter_datasets(datasets, col, out_type=dict)
            json.dump(out_dataset, outfile)

    # show task status
    def show(self, args):
        params = vars(args)
        output = params.pop('output', None)
        outcol = params.pop('outcol', None)
        print('INFO: Fetching PanDA Tasks...')
        datasets = self.query_tasks(**params)
        
        print('')
        self.print_standard(datasets)
        
        if args.metadata:
            print('')
            self.print_metadata(datasets)       

        if output:
            self.save_datasets(datasets, output, col=outcol)


    def get_active_tasks(self):
        return pbookCore.get_active_tasks()

    def kill(self, args):
        task_id = int(args.jeditaskid)
        self.pbook.kill(task_id)

    def retry(self, args):
        task_id = int(args.jeditaskid)
        self.pbook.retry(task_id)

    def killAndRetry(self, args):
        task_id = int(args.jeditaskid)
        self.pbook.killAndRetry(task_id)



    '''
    def deprecated_query(self, params):
                req = self.get_request(kPanDA_JobURL, params)
        response_status = self.get_response(req)['jobs']
        params['fields'] = 'metastruct'
        req = self.get_request(kPanDA_JobURL, params)
        response_metadata = self.get_response(req)['jobs']
        set_trace()
        response = []
        for (status, metadata) in zip(response_status, response_metadata):
            response.append({**status,**metadata})
    '''


import sys, os
import argparse
import json
import yaml
import shutil

from distutils import dir_util
from datetime import datetime
from json import JSONDecodeError
from pdb import set_trace

from hpogrid.utils import stylus, helper
from hpogrid.components.defaults import *
from hpogrid import ConfigurationBase

kConfigList = ['scripts_path', 'model_config', 'search_space', 'hpo_config', 'grid_config']

class ProjectConfiguration(ConfigurationBase):
    
    config_type = 'project'
    description = 'Manage a project for hyperparamter optimization'
    usage = 'hpogrid project <action> <project_name> [<options>]'
    list_columns = ['Project Title']
    show_columns = ['Attribute', 'Value']  
    json_interpreted = []

    def __init__(self):
        self.project_config = {}
        super().__init__()
        
    @staticmethod
    def get_parser_test():
        parser, subparsers = ConfigurationBase.get_parser_test(
            description='Manage a project for hyperparamter optimization',
            usage='hpogrid project <action> <project_name> [<options>')
        for action in subparsers.choices:
            subparsers.choices[action].usage = subparsers.choices[action].usage.replace('<config_type>', 'project')
        for action in ['create', 'update', 'recreate']:
            parser_action = subparsers.choices.get(action, None)
            if parser_action is None:
                raise RuntimeError('Subparser {} not found'.format(action))
            parser_action.add_argument('name', help= "Name given to the project")
            parser_action.add_argument('-p','--scripts_path', metavar='',
                help='Path to the location of training scripts'
                ' (or the directory containing the training scripts)')
            parser_action.add_argument('-o','--hpo_config', metavar='',
                help='Name of the hpo configuration to use')
            parser_action.add_argument('-g','--grid_config', metavar='',
                help='Name of the grid configuration to use')
            parser_action.add_argument('-m','--model_config', metavar='',
                help='Name of the model configuration to use')
            parser_action.add_argument('-s','--search_space', metavar='',
                help='Name of the search space configuration to use')
        return parser

    def get_parser(self, action=None):
        parser = self.get_base_parser()           
        if action in kConfigAction:          
            parser.add_argument('name', help= "Name given the project")
            parser.add_argument('-p','--scripts_path', metavar='',
                help='Path to where the training scripts'
                ' (or the directory containing the training scripts) are located')
            parser.add_argument('-o','--hpo_config', metavar='',
                help='Name of the hpo configuration to use')
            parser.add_argument('-g','--grid_config', metavar='',
                help='Name of the grid configuration to use')
            parser.add_argument('-m','--model_config', metavar='',
                help='Name of the model configuration to use')
            parser.add_argument('-s','--search_space', metavar='',
                help='Name of the search space configuration to use')
        else:
            parser = super().get_parser(action)
        return parser

    def get_updated_config(self, config):
        return config

    def process_config(self, config):
        self.project_config = config

        print('INFO: Loading configurations...')
        # check if path to training scripts exists
        if (config['scripts_path'] is not None):
            scripts_path = config['scripts_path']
            if not os.path.exists(scripts_path):
                print('ERROR: Path to training scripts {} does not exist.'
                       'Copy to project will be skipped.'.format(scripts_path))
                config['scripts_path'] = None
        else:
            print('INFO: Path to training scripts is not specified. Skipping...')

        config_type_map = {
            'hpo_config': 'hpo',
            'grid_config': 'grid',
            'model_config': 'model',
            'search_space': 'search_space'
        }

        # check if input configuration files exist
        for config_type in config_type_map:
            if (config_type in config) and (config[config_type] is not None):
                config_dir = helper.get_config_dir(config_type_map[config_type])
                config_base_name = '{}.json'.format(config[config_type])
                config_path = os.path.join(config_dir, config_base_name)
                config_name = config_type_map[config_type].replace('_',' ')
                if not os.path.exists(config_path):
                    raise FileNotFoundError('Path to {} configuration {}'
                                            ' does not exist.'.format(config_name, config_path))
                with open(config_path,'r') as config_file:
                    config[config_type] = json.load(config_file)
                print('INFO: Loaded {} configuration from {} '.format(config_name, config_path))
            else:
                print('INFO: Path to {} configuration is not specified. Skipping...'.format(config_name))

        return config
    


    @classmethod
    def save(cls, config, name=None, action='create'):
        if action not in kConfigAction:
            raise ValueError('Unknown action "{}"'.format(action))
        proj_name = config.get('project_name', name)
        if not proj_name:
            raise ValueError('Project name not specified in configuration')
        proj_path = cls.get_project_path(proj_name)
        
        action_map = {
            'create': 'Creating',
            'recreate': 'Recreating',
            'update': 'Updating'
        }
        
        print('INFO: {} project "{}"'.format(action_map.get(action, action), proj_name))
        if (os.path.exists(proj_path)):
            if  action == 'create':
                print('ERROR: Project "{}" already exists. If you want to overwrite,'
                    ' use "recreate" or "update" action instead of "create".'.format(proj_name))
                return None
            elif action == 'recreate':
                backup_dir = cls.get_project_path('backup')
                os.makedirs(backup_dir, exist_ok=True)
                timestamp = datetime.now().strftime("%Y%m%d%H%M%S")
                backup_proj_name = os.path.join(backup_dir, '{}_{}'.format(proj_name, timestamp))
                shutil.move(proj_path, backup_proj_name)
                print('INFO: Recreating project. Original project moved to backup directory {}.'.format(
                    backup_proj_name))
        # create project directories
        scripts_dir = os.path.join(proj_path, 'scripts')
        config_dir = os.path.join(proj_path, 'config')
        os.makedirs(proj_path, exist_ok=True)        
        os.makedirs(scripts_dir, exist_ok=True)
        os.makedirs(config_dir, exist_ok=True)

        # copy training scripts to the project directory
        if ('scripts_path' in config) and (config['scripts_path'] is not None):
            print('INFO: Copying traning scripts from {} to {}'.format(config['scripts_path'], scripts_dir))
            # copy contents of directory to project/scrsipts/
            if os.path.isdir(config['scripts_path']):
                helper.copytree(config['scripts_path'], scripts_dir)
            else:
                shutil.copy2(config['scripts_path'], scripts_dir)

        
        project_config = {}
        project_config['project_name'] = proj_name
        
        if action == 'update':
            project_config.update(helper.get_project_config(proj_name))
            
        for key in kConfigList:
            if (key in config) and (config[key] is not None):
                project_config[key] = config[key]
    
        project_config_path_json = os.path.join(config_dir, kProjectConfigNameJson)
        project_config_path_yaml = os.path.join(config_dir, kProjectConfigNameYaml)
        
        with open(project_config_path_json,'w') as proj_config_file:
            json.dump(project_config, proj_config_file, indent=2)
            print('INFO: Created project configuration: {}'.format(project_config_path_json))
        with open(project_config_path_yaml,'w') as proj_config_file:
            yaml.dump(project_config, proj_config_file, default_flow_style=False, sort_keys=False)
            print('INFO: Created project configuration: {}'.format(project_config_path_yaml))
            
        action_map = {
            'create': 'created',
            'recreate': 'recreated',
            'update': 'updated'
        }
        print('INFO: Successfully {} project "{}"'.format(action_map.get(action, action), proj_name))
        
    @classmethod
    def remove(cls, name):
        project_path = cls.get_project_path(name)
        if os.path.exists(project_path):
            print('WARNING: To avoid accidental deletion of important files. '
                'Please delete your project manually at:\n{}'.format(project_path))
        else:
            print('ERROR: Cannot remove project in {}. Path does not exist.'.format(project_path))
            
    @classmethod
    def get_project_path(cls, name):
        """Returns the path to the project
        
        Args:
            name: str
                Name of project
        """
        return helper.get_project_path(name)
    
    @classmethod
    def get_config_path(cls, config_name:str=None):
        """Returns the full path of a project configuration file
        
        Args:
            config_name: str
                Name of the project
        """
        return helper.get_project_config_path(proj_name=config_name)         

    @classmethod
    def list(cls, expr:str=None):
        """List out configuration files for a specific type of configuration as a table
        
        Args:
            expr: str
                Regular expression for filtering name of configuration files        
            exclude: list[str]
                Configuration files to exclude from listing
        """
        project_list = helper.get_project_list(expr)
        table = stylus.create_table(project_list, cls.list_columns)
        print(table)
        
    @classmethod
    def show(cls, name:str):
        """Display the content of a configuration file
        
        Args:
            name: str
                Name of configuration file        
        """
        config = cls.load(name)
        print(yaml.dump(config, allow_unicode=True, default_flow_style=False, sort_keys=False))   


import os
import sys
import json
import yaml
import shutil
import inspect
import tarfile
import fnmatch
import argparse
import numpy as np
import multiprocessing
from typing import List, Dict
from contextlib import contextmanager
from subprocess import Popen, PIPE
import shlex


from hpogrid.components.defaults import *  
from hpogrid.components.environment_settings import *

def run_command(cmd):
    args = shlex.split(cmd, posix=False)
    with Popen(args, stdout=PIPE, bufsize=1, universal_newlines=True) as p:
        while True:
            line = p.stdout.readline()
            if not line:
                break
            print(line)    
        exit_code = p.poll()
    return exit_code


class NpEncoder(json.JSONEncoder):
    def default(self, obj):
        if isinstance(obj, np.integer):
            return int(obj)
        elif isinstance(obj, np.floating):
            return float(obj)
        elif isinstance(obj, np.ndarray):
            return obj.tolist()
        else:
            return super(NpEncoder, self).default(obj)

def copytree(src, dst, symlinks=False, ignore=None):
    for item in os.listdir(src):
        s = os.path.join(src, item)
        d = os.path.join(dst, item)
        if os.path.isdir(s):
            shutil.copytree(s, d, symlinks, ignore)
        else:
            shutil.copy2(s, d)
###############################################################################################
#################### Project Configuration Related         ####################################
###############################################################################################

def set_base_path(path:str) -> None:
    os.environ[kHPOGridEnvPath] = path
    print('INFO: Environment variable {} is set to {}'.format(kHPOGridEnvPath, path))    

def get_base_path():
    if kHPOGridEnvPath not in os.environ:
        print('WARNING: {} environment variable not set.'
            'Will set to current working directory by default.'.format(kHPOGridEnvPath))
        set_base_path(os.getcwd())
    return os.environ.get(kHPOGridEnvPath, os.getcwd())         

def get_config_dir(config_type:str):
    """Returns the directory that stores configuration files of a particular configuration type
    
    Args:
        config_type:str
            Type of configuration
    """
    config_dir = os.path.join(get_base_path(), kConfigPaths[config_type])
    return config_dir
    
def get_project_path(proj_name:str):
    if get_run_mode() in ['GRID', 'IDDS']:
        proj_path = get_workdir()
    elif get_run_mode() == 'LOCAL':
        proj_dir = get_config_dir('project')
        proj_path = os.path.join(proj_dir, proj_name)
    return proj_path

def get_project_config_path(proj_name:str):
    project_path = get_project_path(proj_name)
    config_path = os.path.join(project_path, 'config', kProjectConfigNameJson)
    return config_path
                        
def get_project_config(proj_name:str):
    config_path = get_project_config_path(proj_name)
    if not os.path.exists(config_path):
        raise FileNotFoundError('Missing project configuration file: {}'.format(config_path))
    config = load_configuration(config_path)
    return config     

def get_project_list(expr=None):
    if get_run_mode() in ['GRID', 'IDDS']:
        config_path = os.path.join(get_workdir(), 'config', kProjectConfigNameJson)
        config = load_configuration(config_path)
        project_list = [config['project_name']]
    elif get_run_mode() == 'LOCAL':
        base_path = get_base_path()
        proj_dir = os.path.join(base_path, kConfigPaths['project'])
        project_list = [d for d in os.listdir(proj_dir) if 
                        os.path.isdir(os.path.join(proj_dir, d))]
    expr = expr if expr is not None else '*'
    project_list = fnmatch.filter(project_list, expr)
    project_list = [p for p in project_list if p!='backup']

    return project_list

def is_project(obj:str):
    if not isinstance(obj, str):
        return False
    return obj in get_project_list()      

def load_configuration(config_input:[Dict, str]) -> Dict:
    """Returns project configuration given input
    
    Args:
        config_input: dict, str
            If dict, it is the configuration itself
            If str, it is either the name of the project,
            of path to a configuration file.
    """
    if isinstance(config_input, dict):
        return config_input
    elif isinstance(config_input, str):
        ext = os.path.splitext(config_input)[1]
        # check if config_input is a project name
        if not ext:
            if is_project(config_input):
                return get_project_config(config_input)
            else:
                raise ValueError('Project "{}" does not exist'.format(config_input))
        # check if configuration file exists
        if not (os.path.exists(config_input)):
            raise FileNotFoundError("Configuration file {} does not exist.".format(config_input))
        # load configuration file
        with open(config_input, 'r') as file:
            if ext in ['.txt', '.yaml']:
                config = yaml.safe_load(file)
            elif ext == '.json':
                config = json.load(file)
            else:
                raise ValueErrror('The configuration file has an unsupported '
                              'file extension: {}\n Supported file extensions '
                              'are .txt, .yaml or .json'.format(ext))
        return config
    else:
        raise ValueError('Unknown configuration format: {}'.format(type(config_input)))

        
def set_scripts_path(scripts_path, undo=False):
    if (scripts_path in sys.path) and (undo==True):
        print('INFO: Removed {} from $PYTHONPATH'.format(scripts_path))
        sys.path.remove(scripts_path)
        os.environ["PYTHONPATH"].replace(scripts_path+":","")
        
    if (scripts_path not in sys.path) and (undo==False):
        print('INFO: Adding {} to $PYTHONPATH'.format(scripts_path))
        sys.path.append(scripts_path)
        os.environ["PYTHONPATH"] = scripts_path + ":" + os.environ.get("PYTHONPATH", "")

def get_scripts_path(proj_name):
    project_path = get_project_path(proj_name)
    scripts_path = os.path.join(project_path, 'scripts')
    return scripts_path
     
        
def set_scripts_path_from_project(proj_name, undo=False):
    project_path = get_project_path(proj_name)
    scripts_path = os.path.join(project_path, 'scripts')
    set_scripts_path(scripts_path, undo=undo)   
    
################################################################################################ 
################################################################################################ 

def pretty_path(path:str):
    return os.path.abspath(os.path.realpath(os.path.expandvars(os.path.expanduser(path))))

def pretty_dirname(path:str):
    path = pretty_path(path)
    if os.path.isdir(path):
        return path
    else:
        return os.path.dirname(path)
        
@contextmanager
def cd(newdir):
    prevdir = os.getcwd()
    os.chdir(os.path.expanduser(newdir))
    try:
        yield
    finally:
        os.chdir(prevdir)

def get_physical_devices(device_type='GPU'):
    import tensorflow as tf
    physical_devices = tf.config.list_physical_devices('GPU')
    return physical_devices

def get_n_gpu():
    import torch
    return torch.cuda.device_count()

def get_n_cpu():
    return multiprocessing.cpu_count()


def extract_tarball(in_path:str, out_path:str) ->List[str]:
    tarfiles = [ f for f in os.listdir(in_path) if f.endswith('tar.gz')]
    extracted_files = []
    for f in tarfiles:
        tar = tarfile.open(f, "r:gz")
        print('INFO: Untaring the file {}'.format(f))
        tar.extractall(path=out_path)
        extracted_files += tar.getnames()
        tar.close()
    return extracted_files

def remove_files(files:List[str]):
    for f in files:
        if os.path.isfile(f):
            os.remove(f)
        elif os.path.isdir(f):
            shutil.rmtree(f)        
            
def is_function_or_method(obj):
    """Check if an object is a function or method.
    Args:
        obj: The Python object in question.
    Returns:
        True if the object is an function or method.
    """
    return inspect.isfunction(obj) or inspect.ismethod(obj) or is_cython(obj)


def is_class_method(f):
    """Returns whether the given method is a class_method."""
    return hasattr(f, "__self__") and f.__self__ is not None  


"""
import socket
hostname = socket.gethostname()
"""



from past.utils import old_div
import numpy as np

from ray import tune

from hpogrid.search_space.base_space import BaseSpace

class PBTSpace(BaseSpace):

	def __init__(self, search_space = None):
		self.library = 'pbt'
		super().__init__(search_space)	
		

	def reset_space(self):
		self.search_space = {}

	def append(self, space_value):
		self.search_space[space_value[0]] = space_value[1]

	def categorical(self, label, categories, grid_search = False):
		return (label, categories)

	def uniform(self, label, low, high):
		return (label, lambda: np.random.uniform(low, high))
	
	def uniformint(self, label, low, high):
		return (label, lambda : np.random.randint(low, high))

	def loguniform(self, label, low, high, base=10):
		logmin = np.log(low) / np.log(base)
		logmax = np.log(high) / np.log(base)
		return (label, lambda: base**(np.random.uniform(logmin, logmax)))

	def qloguniform(self, label, low, high, base=10):
		logmin = np.log(low) / np.log(base)
		logmax = np.log(high) / np.log(base)
		return (label, lambda: 
			np.round(old_div(base**(np.random.uniform(logmin, logmax)),q))*q)

	def quniform(self, label, low, high, q):
		return (labelm, lambda: 
			np.round(old_div(np.random.uniform(low, high),q))*q)

	def quniformint(self, label, low, high, q):
		return (label, lambda : 
			np.round(old_div(np.random.uniform(low, high),q))*q)

	def normal(self, mu, sigma):
		return (label, lambda : sigma*np.random.randn()+mu)

	def qnormal(self, mu, sigma, q):
		return (label, lambda : 
			np.round(old_div(sigma*np.random.randn()+mu,q))*q)

	def lognormal(self, mu, sigma, base = np.e):
		return (label, lambda : 
			np.power(sigma*np.random.randn()+mu, base))

	def qlognormal(self, mu, sigma, q, base = np.e):
		return (label, lambda _: 
			np.round(old_div(np.power(sigma*np.random.randn()+mu, base),q))*q )	

	def fixed(self, value):
		return (label, [value])


import os
import argparse
from pdb import set_trace
import tarfile

_help_test = 'test'

kDefaultDisk = 'UKI-NORTHGRID-MAN-HEP_SCRATCHDISK'
kDefaultDatasetName = 'user.${RUCIO_ACCOUNT}.hpo.test.dataset.01'
kDefaultTempDir = './'

def get_rucio_parser():
	
	parser = argparse.ArgumentParser()
	parser.add_argument('data_paths', nargs='+', help=_help_test)
	parser.add_argument('-n', '--name', help=_help_test, default=kDefaultDatasetName)
	parser.add_argument('-t', '--temp_dir', help=_help_test, default=kDefaultTempDir)
	parser.add_argument('-d','--disk', help=_help_test, default=kDefaultDisk)
	parser.add_argument('-s', '--save', action='store_true', help=_help_test, default=False)

	return parser


def upload_rucio(args):
	disk_name = args.disk
	dataset_name = args.name
	tarfile_name = args.name.replace('user.${RUCIO_ACCOUNT}.','')+'.tar.gz'
	tarfile_fullpath = os.path.join(args.temp_dir, tarfile_name)
	make_tarfile(args.data_paths, tarfile_fullpath)

	os.system('rucio add-dataset {}'.format(dataset_name))
	os.system('rucio upload {} --name {} --rse {}'.format(tarfile_fullpath, tarfile_name, disk_name))
	os.system('rucio attach {} user.${{RUCIO_ACCOUNT}}:{}'.format(dataset_name, tarfile_name))


def make_tarfile(source, outname):
	print('INFO: Making tarfile for {}'.format(source))
    with tarfile.open(outname, "w:gz") as tar:
    	for src in source:
    		src_path = os.path.abspath(os.path.expanduser(src.rstrip('/')))
        	tar.add(src_path, arcname=os.path.basename(src_path))

if __name__ == '__main__':
	parser = get_rucio_parser()
	args = parser.parse_args()
	upload_rucio(args)



import os
import sys
import glob
import argparse
import json
import tempfile

import hpogrid
from hpogrid.components.defaults import *
from hpogrid.utils import helper, stylus
from hpogrid.idds_interface.idds_utils import transpose_config
from hpogrid.components.validation import validate_project_config

class GridHandler():
    def __init__(self):

        # submit grid job via hpogrid executable
        if len(sys.argv) > 1:
            self.run_parser()

    def get_parser(self):
        parser = argparse.ArgumentParser(
                    formatter_class=argparse.RawDescriptionHelpFormatter)
        parser.add_argument('input', help='the name of project or configuration'
                            ' file for grid job submission')               
        parser.add_argument('-n','--n_jobs', type=int, help='number of jobs to submit',
            default=1)
        parser.add_argument('-s','--site', help='site to submit the job to '
            '(this will override the grid config site setting)', choices=kGPUGridSiteList)
        parser.add_argument('-t','--time', help='same as maxCpuCount in prun which '
            'specifies the maximum cpu time for a job (prevent being killed by'
            'looping job detection)', type=int, default=-1)
        parser.add_argument('-m','--mode', help='mode for job submission (choose between '
                            '"grid" or "idds"',
                            choices=['grid','idds'],
                            default='grid')
        parser.add_argument('-p','--search_points', help='json file containing a list of '
                            'search points to evaluate (for non-idds jobs only)', default=None)
        return parser

    def run_parser(self):
        parser = self.get_parser()
        if os.path.basename(sys.argv[0]) == 'hpogrid':
            args, extra = parser.parse_known_args(sys.argv[2:])
        else:
            args, extra = parser.parse_known_args(sys.argv[1:])
        self.submit(args.input, 
                    n_jobs=args.n_jobs,
                    site=args.site,
                    time=args.time,
                    search_points = args.search_points,
                    mode=args.mode,
                    extra=extra)
    
    @staticmethod
    def submit(config_input:str, n_jobs=1, site=None, time=-1, search_points=None, mode='grid', extra=[]):
        config = GridHandler.preprocess_config(config_input)
        if mode == 'grid':
            GridHandler.submit_grid(config, n_jobs=n_jobs, site=site,
                                    time=time, search_points=search_points, extra=extra)
        elif mode == 'idds':
            GridHandler.submit_idds(config, n_jobs=n_jobs, site=site, extra=extra)
        else:
            raise ValueError('Unknown job submission mode: {}'.format(mode))
            
    @staticmethod
    def preprocess_config(config_input):
        config = hpogrid.load_configuration(config_input)
        config = validate_project_config(config)
        project_name = config['project_name']
        if not helper.is_project(project_name):
            print('INFO: Creating project directory for "{}"'.format(project_name))
            hpogrid.create_project(config, 'create')
        elif not helper.is_project(config_input):
            print('INFO: Replacing project directory for "{}"'.format(project_name))
            hpogrid.create_project(config, 'recreate')
        return config
        
    @staticmethod
    def submit_idds(config, n_jobs=1, site=None, extra=[]):
        project_path = helper.get_project_path(config['project_name'])
        print('INFO: Submitting {} idds grid job(s)'.format(n_jobs))
        with tempfile.TemporaryDirectory() as tmpdirname:
            idds_config, search_space = transpose_config(config)
            idds_config_path = os.path.join(tmpdirname, 'idds_config.json')
            search_space_path = os.path.join(tmpdirname, 'search_space.json')
            # create temporary files for idds configuration and search space
            with open(idds_config_path, 'w') as idds_config_file:
                json.dump(idds_config, idds_config_file)
            with open(search_space_path, 'w') as search_space_file:
                json.dump(search_space, search_space_file)  
            command = GridHandler.format_idds_command(config, 
                                               tmpdirname,
                                               site=site,
                                               extra=extra)
            with helper.cd(project_path):
                # submit jobs
                for _ in range(n_jobs):
                    os.system(command)
    @staticmethod
    def submit_grid(config, n_jobs=1, site=None, time=-1, search_points=None, extra=[]):
        project_path = helper.get_project_path(config['project_name'])
        print('INFO: Submitting {} grid job(s)'.format(n_jobs))
        command = GridHandler.format_grid_command(config, site=site,
                                                  time=time,
                                                  search_points=search_points,
                                                  extra=extra)
        with helper.cd(project_path):
            # submit jobs
            for _ in range(n_jobs):
                os.system(command)
                
    @staticmethod
    def format_idds_command(config, tmpdirname, site=None, extra=[]):
        idds_config_path = os.path.join(tmpdirname, 'idds_config.json')
        search_space_path = os.path.join(tmpdirname, 'search_space.json')
        options = {'loadJson': idds_config_path,
                   'searchSpaceFile': search_space_path}
        proj_name = config['project_name']
        grid_config = config['grid_config']
        # override site settings if given
        if not site:
            site = grid_config['site']
                
        if (site != 'ANY'):
            options['site'] = site
            # specify architecture for gpu/cpu site
            if 'GPU' in site:
                options['architecture'] = 'nvidia-gpu'
        # options excluded from idds configuration file due to
        # possible bash variable expansion
        in_ds = config['grid_config']['inDS']
        out_ds = config['grid_config']['outDS']
        
        if in_ds:
            options['trainingDS'] = in_ds
            
        if '{HPO_PROJECT_NAME}' in out_ds:
            out_ds = out_ds.format(HPO_PROJECT_NAME=proj_name)
        options['outDS'] = out_ds
                
        command = 'phpo {} {}'.format(stylus.join_options(options), ' '.join(extra))    
        return command
    
    @staticmethod
    def format_grid_command(config, site=None, time=-1, search_points=None, extra=[]):
        options = {'forceStaged': '',
                   'useSandbox': '',
                   'noBuild': '',
                   'alrb': ''}
        
        grid_config = config['grid_config']
        proj_name = config['project_name']
        
        options['containerImage'] = grid_config['container']
        
        search_points = '' if not search_points else '--search_points {}'.format(search_points)

        options['exec'] = '"pip install --upgrade hpogrid && '+\
                          'hpogrid run {} --mode grid {}"'.format(proj_name, search_points)

        if not grid_config['retry']:
            options['disableAutoRetry'] = ''

        if grid_config['inDS']:
            options['inDS'] = grid_config['inDS']

        if '{HPO_PROJECT_NAME}' in grid_config['outDS']:
            grid_config['outDS'] = grid_config['outDS'].format(HPO_PROJECT_NAME=proj_name)
                
        options['outDS'] = grid_config['outDS']
                
        if time != -1:
            options['maxCpuCount'] = str(time)
        
        # override site settings if given
        if not site:
            site = grid_config['site']
                
        if (site != 'ANY'):
            options['site'] = site
            # specify architecture for gpu/cpu site
            if 'GPU' in site:
                options['cmtConfig'] = 'nvidia-gpu'
                
        
        command = 'prun {} {}'.format(stylus.join_options(options), ' '.join(extra))
        return command        
                


import os
import sys
import glob
import argparse
import json
from typing import Optional
import tempfile
from hpogrid.components.defaults import *
from hpogrid.utils import helper, stylus
from hpogrid.components import environment_settings
from hpogrid.components.validation import validate_project_config

def transpose_config(config):
    project_name = config['project_name']
    evaluation_container = config['grid_config']['container']
    search_space = config['search_space']
    library = config['hpo_config']['algorithm']
    method = config['hpo_config']['algorithm_param'].get('method', None)
    max_point = config['hpo_config']['num_trials']
    max_concurrent = config['hpo_config']['max_concurrent']


    # format generator command
    generator_options = {}
    generator_options['n_point'] = '%NUM_POINTS'
    generator_options['max_point'] = '%MAX_POINTS'
    generator_options['infile'] = os.path.join(kiDDSBasePath, '%IN')
    generator_options['outfile'] = os.path.join(kiDDSBasePath, '%OUT')
    generator_options['lib'] = library
    if (library == 'nevergrad') and method:
        generator_options['method'] = method
    generator_cmd = 'hpogrid generate {}'.format(stylus.join_options(generator_options))

    # format idds configuration
    idds_config = {}
    idds_config['steeringExec'] = ("run --rm -v \"$(pwd)\":{idds_base_path} {container} " +\
    "/bin/bash -c \"{cmd}\" ").format(idds_base_path=kiDDSBasePath, container=kDefaultiDDSContainer, cmd=generator_cmd)
    #idds_config['steeringExec'] = ("run --rm -v \"$(pwd)\":{idds_base_path} {container} " +\
    #"/bin/bash -c \"{cmd}\" ").format(idds_base_path="/HPOiDDS", container="gitlab-registry.cern.ch/zhangruihpc/steeringcontainer:0.0.2", cmd=generator_cmd)

    idds_config['evaluationExec'] = "pip install --upgrade hpogrid && " + \
                                    "hpogrid run {project_name} --mode idds".format(project_name=project_name)
    idds_config['evaluationContainer'] = evaluation_container
    idds_config['evaluationInput'] = kiDDSHPinput
    idds_config['evaluationOutput'] = kiDDSHPoutput
    idds_config['evaluationMeta'] = kGridSiteMetadataFileName
    idds_config['nParallelEvaluation'] = max_concurrent
    idds_config['maxPoints'] = max_point
    return idds_config, search_space


def get_search_points():
    if not environment_settings.is_idds_job():
        raise RuntimeError('Cannot get search points for non-idds environment')
    
    workdir = helper.get_workdir()
    search_point_path = os.path.join(workdir, kiDDSHPinput)
    if not os.path.exists(search_point_path):
        raise FileNotFoundError('No such file or directory: \'{}\''.format(search_point_path))
    with open(search_point_path,'r') as search_point_file:
        search_points = json.load(search_point_file)
    return search_points

def create_idds_output_from_metadata(metadata):
    idds_output = {}
    idds_output['status'] = 0
    loss = []
    metric = metadata['metric']
    for index in metadata['result']:
        loss.append(metadata['result'][index][metric])
    if len(loss) == 1:
        loss = loss[0]
    idds_output['loss'] = loss
    idds_output['message'] = ''
    return idds_output


def save_idds_output_from_metadata(metadata):
    if not environment_settings.is_idds_job():
        raise RuntimeError('save_output should not be called in non-idds environment')    
        
    output = create_idds_output_from_metadata(metadata)
    workdir = helper.get_workdir()
    output_path = os.path.join(workdir, kiDDSHPoutput)
    with open(output_path, 'w') as output_file:
        print('INFO: Saved idds output at {}'.format(output_path))
        json.dump(output, output_file, cls=helper.NpEncoder)
        
        
        
def check_log(workload_id:int, request_id=None,
              idds_config_path=kDefaultiDDSConfigPath,
              proxy_path=kDefaultProxyPath):
    os.environ['IDDS_CONFIG'] = idds_config_path
    original_proxy = os.environ.get('X509_USER_PROXY', None)
    os.environ['X509_USER_PROXY'] = proxy_path
    from idds.client.client import Client
    from idds.common.utils import get_rest_host
    host = get_rest_host()
    client = Client(host=host)
#    from pdb import set_trace
#    set_trace()
    print('INFO: Fetching iDDS log files for the task: {}'.format(workload_id))
    filename = client.download_logs(workload_id=workload_id, request_id=request_id, dest_dir='/tmp')
    print('INFO: iDDS log files can be found at:')
    print(filename)
    if original_proxy:
        os.environ['X509_USER_PROXY'] = original_proxy
        
        


import os
import sys
import copy
import json
import argparse
from pdb import set_trace

import pandas as pd
from tabulate import tabulate

from hpogrid.components.panda_task_manager import PandaTaskManager
from hpogrid.components.defaults import *
from hpogrid.components.validation import validate_job_metadata

kDefaultExtraColumns = []
kSupportedExtraColumns = ['site', 'task_time_s', 'time_s', 'taskid', 'start_timestamp',
                         'start_datetime','end_datetime']


class HPOReport():
    def __init__(self, data=None, verbose=True):
        self.verbose = verbose
        self.reset()
        if data:
            self.append(data)


    def reset(self):
        self.attrib = {}
        self.data = []

    def set_primary(self, data):
        self.attrib['project_name'] = data['project_name']
        self.attrib['hyperparameters'] = data['hyperparameters']
        self.attrib['metric'] = data['metric']
        self.attrib['mode'] = data['mode']
        if self.verbose:
            print('INFO: Setting up HPO report with the following attributes')
            print('Project Name     : {project_name} \n'
                  'Hyperparameters  : {hyperparameters}\n'
                  'Metric           : {metric}\n'
                  'Mode             : {mode}'.format(**self.attrib))

    def check_consistency(self, data):
        if not validate_job_metadata(data):
            if self.verbose:
                print('ERROR: Invalid format for HPO result summary. '
                    'Result summary will be skipped')
            return False

        if not self.data:
            self.set_primary(data)

        if (data['project_name'] != self.attrib['project_name']):
            if self.verbose:
                print('INFO: Expecting HPO project name "{}" but "{}" is received. '
                    'Result summary will be skipped.')
            return False
        ####################################################################################### 
        if (set(self.attrib['hyperparameters']) != set(data['hyperparameters'])):
            if self.verbose:
                print('INFO: Inconsistent hyperparameter space. '
                        'Result summary will be skipped.')
            return False
        #######################################################################################
        if (self.attrib['metric'] != data['metric']) or (self.attrib['mode'] != data['mode']):
            if self.verbose:
                print('INFO: Inconsistent metric definition. '
                    'Result summary will be skipped.')
            return False
        return True

    def append(self, data, extras=None):
        if not isinstance(data, list):
            data = [data]
        for d in data:
            if not self.check_consistency(d):
                return None
            keys_to_save = ['start_datetime', 'start_timestamp',
                            'end_datetime', 'task_time_s', 'best_config', 'result']
            result = {key: d[key] for key in keys_to_save}
            if extras:
                for key in extras:
                    result[key] = extras[key]
            self.data.append(result)

    def _check_data(self):
        if not self.data:
            print('ERROR: No HPO results to report.')
            return False
        return True

    @classmethod
    def from_json(cls, fname, verbose=True):
        with open(fname, 'r') as json_file:
            data = json.load(json_file)
        return cls(data, verbose)
    
    @classmethod
    def from_summary(cls, fname, verbose=True):
        with open(fname, 'r') as json_file:
            data = json.load(json_file)
        instance = cls()
        instance.data = data.pop('jobs', None)
        instance.attrib = data
        return instance

    def merge_data(self, extra_columns=None, skip_time=True,
                   metric_min=None, metric_max=None):
        merged_data = []
        for data in self.data:
            for index in data['result']:
                result = copy.deepcopy(data['result'][index])
                merged_data.append(result)
        if extra_columns:
            extras = []
            for data in self.data: 
                for index in data['result']:
                    extras.append({col: data[col] for col in extra_columns})
            for data, extra in zip(merged_data, extras):
                data.update(extra)
        if skip_time:
            for data in merged_data:
                data.pop('time_s')
        df = pd.DataFrame(merged_data)
        
        if self.attrib['mode'] == 'min':
            sort_ascending = True
        elif self.attrib['mode'] == 'max':
            sort_ascending = False

        df = df.sort_values(by=self.attrib['metric'] , ascending=sort_ascending).reset_index(drop=True)
        if metric_min is not None:
            df = df[df[self.attrib['metric']] > metric_min]
        if metric_max is not None:
            df = df[df[self.attrib['metric']] < metric_max]
        merged_data = df.to_dict('records')
        return merged_data

    def to_summary(self, fname, indent=2):
        if not self._check_data():
            return None
        report = self.attrib
        report['jobs'] = self.data
        with open(fname, 'w') as file:
            json.dump(report, file, indent=indent)

    def to_json(self, fname, extra_columns=None, skip_time=True, 
                metric_min=None, metric_max=None, indent=2):
        if not self._check_data():
            return None
        data = self.merge_data(extra_columns, skip_time, 
                               metric_min=metric_min,
                               metric_max=metric_max)
        with open(fname, 'w') as output:
            json.dump(data, output, indent=indent)

    def to_dataframe(self, extra_columns=None, skip_time=True,
                    metric_min=None, metric_max=None):
        if not self._check_data():
            return None
        data = self.merge_data(extra_columns, skip_time,
                               metric_min=metric_min,
                               metric_max=metric_max)                               
        df = pd.DataFrame(data)
        return df

    def to_mlflow(self, extra_columns=None, skip_time=True,
                 metric_min=None, metric_max=None):
        if not self._check_data():
            return None
        from mlflow.tracking import MlflowClient
        client = MlflowClient()
        experiment_id = client.create_experiment(self.attrib['project_name'])
        hyperparameters = self.attrib['hyperparameters']
        metric = self.attrib['metric']
        
        if not extra_columns:
            extra_columns = ['start_timestamp']
        elif 'start_timestamp' not in extra_columns:
            extra_columns.append('start_timestamp')
        merged_data = self.merge_data(extra_columns, skip_time=False, 
                                       metric_min=metric_min,
                                       metric_max=metric_max)
        for data in merged_data:
            start_time = int(data['start_timestamp']*1000)
            end_time = int((data['start_timestamp']+data['time_s'])*1000)
            run = client.create_run(experiment_id, start_time=start_time)
            run_id = run.info.run_id
            for hp in hyperparameters:
                client.log_param(run_id, hp, data[hp])
            client.log_metric(run_id, metric, data[metric])
            client.set_terminated(run_id, end_time=end_time)
            
    def to_html(self, fname=None, extra_columns=None, skip_time=True,
               metric_min=None, metric_max=None):
        if not self._check_data():
            return None
        df = self.to_dataframe(extra_columns, skip_time,
                               metric_min=metric_min,
                               metric_max=metric_max)                               
        html_text = df.to_html()
        if fname:
            with open(fname, 'w') as outfile:
                outfile.write(html_text)
        return html_text

    def to_parallel_coordinate_plot(self, fname=None, extra_columns=None, skip_time=True,
                                   metric_min=None, metric_max=None):
        if not self._check_data():
            return None
        import hiplot as hip
        merged_data = self.merge_data(extra_columns, skip_time, 
                                       metric_min=metric_min,
                                       metric_max=metric_max)
        html_text = hip.Experiment.from_iterable(merged_data).to_html()
        if fname:
            with open(fname, 'w') as outfile:
                outfile.write(html_text)
        return html_text

    def to_csv(self, fname, extra_columns=None, skip_time=True,
               metric_min=None, metric_max=None):
        if not self._check_data():
            return None
        df = self.to_dataframe(extra_columns, skip_time, 
                               metric_min=metric_min,
                               metric_max=metric_max)
        csv = df.to_csv(fname, encoding='utf-8')
        return csv

    def show(self, extra_columns=None, skip_time=True):
        if not self._check_data():
            return None
        df = self.to_dataframe(extra_columns, skip_time)
        print(tabulate(df, showindex=True, headers=df.columns, tablefmt="psql",stralign='center'))


class HPOTaskHandle(HPOReport):

    def __init__(self):
        super().__init__()
        self.taskmgr = PandaTaskManager()
        if len(sys.argv) > 1:
            self.run_parser()

    def get_parser(self, method=None):
        parser = argparse.ArgumentParser(description='Tool for managing HPO result',
                    formatter_class=argparse.RawDescriptionHelpFormatter)    
        parser.add_argument('proj_name', help='Name of the project')
        parser.add_argument('-l','--limit', type=int, default=1000,
            help='the maximum number of tasks to query')
        parser.add_argument('-d', '--days', type=int, default=30,
            help='filter tasks within the recent N days')
        parser.add_argument('-n', '--taskname',
            help='filter tasks by taskname (accept wildcards)')
        parser.add_argument('-r', '--range',
            help='filter tasks by jeditaskid range')
        parser.add_argument('-e', '--extra', nargs='+',
            default=kDefaultExtraColumns,
            help='extra data columns to be displayed and saved')
        parser.add_argument('-s', '--summary', action='store_true',
            help='save a complete hpo result summary as a json file')        
        parser.add_argument('-j', '--to_json', action='store_true',
            help='output result to a json file')
        parser.add_argument('-t', '--to_html', action='store_true',
            help='output result to an html file')
        parser.add_argument('-c', '--to_csv', action='store_true',
            help='output result to a csv file')
        parser.add_argument('-p', '--to_pcp', action='store_true',
            help='output result as a parallel coordinate plot')
        parser.add_argument('-f', '--to_mlflow', action='store_true',
            help='output result as an mlflow tracker directory')
        parser.add_argument('-o', '--outname', default='hpo_result',
            help='output file name (excluding extension)')
        parser.add_argument('--min', type=float, default=None,
            help='filter results by minimum value of metric ')
        parser.add_argument('--max', type=float, default=None,
            help='filter results by maximum value of metric ')
        return parser

    def run_parser(self):
        parser = self.get_parser()
        args = parser.parse_args(sys.argv[2:])
        self.display(args)

    def display(self, args):
        params = vars(args)
        params['metadata'] = True
        params['status'] = ['finished', 'done']
        proj_name = params.pop('proj_name', '')
        outname = params.pop('outname', None)
        datasets = self.taskmgr.query_tasks(**params)
        # filter tasks by jetitaskid range
        if args.range is not None:
            datasets = self.taskmgr.filter_range(datasets, args.range)
        col = ['computingsite', 'jobs_metadata', 'jeditaskid', 'site']
        datasets = self.taskmgr.filter_datasets(datasets, col)

        extra_columns = args.extra or []
        for extra in args.extra:
            if not extra in kSupportedExtraColumns:
                raise ValueError('Unsupported extra column {}'.format(extra))

        skip_time = 'tims_s' not in extra_columns
        self.parse_datasets(datasets, proj_name, extra_columns=extra_columns)
        self.show(extra_columns, skip_time)

        bounds = {'metric_min':args.min, 'metric_max':args.max}
        
        if args.summary:
            self.to_summary(outname+'_summary.json')
        if args.to_json:
            self.to_json(outname+'.json', extra_columns, skip_time, **bounds)
        if args.to_html:
            self.to_html(outname+'.html', extra_columns, skip_time, **bounds)
        if args.to_csv:
            self.to_csv(outname+'.csv', extra_columns, skip_time, **bounds)    
        if args.to_pcp:
            self.to_parallel_coordinate_plot(outname+'_parallel_coordinate_plot.html',
                                             extra_columns, skip_time, **bounds)   
        if args.to_mlflow:
            self.to_mlflow(extra_columns, skip_time, **bounds)    

    def filter_invalid_datasets(self, datasets, proj_name):
        filtered_datasets = []
        metadata_list = []
        for dataset in datasets:
            metadata = self.extract_metadata(dataset)
            # check if metadata is empty
            if not metadata:
                continue
            # check if it is a valid metadata generated by hpogrid
            # and if the metadata corresponds to the project of concern
            if not all( validate_job_metadata(data) and 
                       data.get('project_name', None) == proj_name \
                       for data in metadata):
                continue
            filtered_datasets.append(dataset)
            metadata_list += metadata
        if len(filtered_datasets) == 0:
            print('INFO: No results found.')
            sys.exit(1)
        metrics = set([metadata['metric'] for metadata in metadata_list])
        if not len(metrics)==1 :
            raise RuntimeError('HPO results from different tasks consist of different metrics: {}'.format(metrics))
        hparams = [metadata['hyperparameters'] for metadata in metadata_list]
        if not all(hp == hparams[0] for hp in hparams):
            raise RuntimeError('HPO results from different tasks consist of different hyperparameter space')
        return filtered_datasets

    def extract_metadata(self, dataset):
        metadata_list = []
        if 'jobs_metadata' in dataset:
            jobs_metadata = dataset['jobs_metadata']
            # take care of general case where a job contains multiple sub-jobs
            # which will return a list of metadata
            if not isinstance(jobs_metadata, list):
                jobs_metadata = [jobs_metadata]
            for job_metadata in jobs_metadata:
                if not isinstance(job_metadata, dict):
                    raise ValueError('Invalid job metadata format')
                keys = list(job_metadata.keys())
                # skip empty metadata
                if not keys:
                    continue
                metadata = job_metadata[keys[0]].get('user_job_metadata', None)
                if metadata is not None:
                    metadata_list.append(metadata)
                else:
                    print('WARNING: The key "user_job_metadata" not found in job metadata. '
                          'Probably the format has changed. Skipping.')
        return metadata_list

    def parse_datasets(self, datasets, proj_name, extra_columns=kDefaultExtraColumns):
        summary = []
        valid_datasets = self.filter_invalid_datasets(datasets, proj_name)

        metric =  None
        mode = None
        
        for dataset in valid_datasets:
            metadata = self.extract_metadata(dataset)
            extras = {}
            if 'taskid' in extra_columns:
                extras['taskid'] = dataset['jeditaskid']
            if 'site' in extra_columns:
                if 'computingsite' in dataset:
                    extras['site'] = dataset['computingsite']
                elif 'site' in dataset:
                    extras['site'] = dataset['site']
                else:
                    extras['site'] = ''
            self.append(metadata, extras)



import sys, os
import argparse
import json

from hpogrid.components.defaults import *
from hpogrid import ConfigurationBase

class ModelConfiguration(ConfigurationBase):

    config_type = 'model'
    description = 'Manage configuration for machine learning model'
    usage = 'hpogrid model_config <action> <config_name> [<options>]'
    list_columns = ['Model Configuration']
    show_columns = ['Attribute', 'Value']  
    json_interpreted = ['param']
    
    @staticmethod
    def get_parser_test():
        parser, subparsers = ConfigurationBase.get_parser_test(
            description='Manage configuration for the machine learning model',
            usage='hpogrid model_config <action> <config_name> [<options>')
        for action in subparsers.choices:
            subparsers.choices[action].usage = subparsers.choices[action].usage.replace('<config_type>', 'model_config')
        for action in ['create', 'update', 'recreate']:
            parser_action = subparsers.choices.get(action, None)
            if parser_action is None:
                raise RuntimeError('Subparser {} not found'.format(action))
            parser_action.add_argument('name', help= "Name given to the configuration file")            
            parser_action.add_argument('-s','--script', metavar='',
                help='Name of the training script where the function or class that defines'
                     ' the training model will be called to perform the training')
            parser_action.add_argument('-m','--model', metavar='',
                help='Name of the function or class that defines the training model')        
            parser_action.add_argument('-p','--param', metavar='',
                help='Extra parameters to be passed to the training model',
                default=kDefaultModelParam)
        return parser

    def get_parser(self, action=None):
        parser = self.get_base_parser()              
        if action in kConfigAction:  
            parser.add_argument('name', help= "Name given to the configuration file")            
            parser.add_argument('-s','--script', metavar='',
                help='Name of the training script where the function or class that defines'
                     ' the training model will be called to perform the training')
            parser.add_argument('-m','--model', metavar='',
                help='Name of the function or class that defines the training model')        
            parser.add_argument('-p','--param', metavar='',
                help='Extra parameters to be passed to the training model',
                default=kDefaultModelParam)
        else:
            parser = super().get_parser(action)
        return parser



from ax.service.ax_client import AxClient

import ray
from ray.tune.suggest.ax import AxSearch

from hpogrid.search_space.ax_space import AxSpace

from pdb import set_trace

class AxAlgoWrapper():
    def __init__(self):
        self.algorithm = None
    def create(self, metric, mode, search_space, **args):
        ax_client = AxClient(enforce_sequential_optimization=False, verbose_logging=True)
        search_space = AxSpace(search_space).get_search_space()

        if mode == 'max':
            minimize = False
        elif mode == 'min':
            minimize = True
        else:
            raise ValueError('mode of evaluation metric can only be "min" or "max"')

        ax_client.create_experiment(
            parameters=search_space,
            objective_name=metric,
            minimize=minimize)

        self.algorithm = AxSearch(ax_client, **args)
        return self.algorithm


import sys, os
import argparse
import json
from json import JSONDecodeError

from hpogrid.components.defaults import *
from hpogrid import ConfigurationBase

class HPOConfiguration(ConfigurationBase):
    
    config_type = 'hpo'
    description = 'Manage configuration for hyperparameter optimization'
    usage = 'hpogrid hpo_config <action> <config_name> [<options>]'
    list_columns = ['HPO Configuration']
    show_columns = ['Attribute', 'Value']
    json_interpreted = ['stop', 'scheduler_param', 'algorithm_param', 'resource']
    
    @staticmethod
    def get_parser_test():
        parent_parser = argparse.ArgumentParser(add_help=False)
        parent_parser.add_argument('name', help= "Name given to the configuration file")
        parent_parser.add_argument('-a','--algorithm', metavar='',
                            help='Algorithm for hyperparameter optimization. '+\
                                 'Allowed values: {}'.format(','.join(kSearchAlgorithms)), 
                            default=kDefaultSearchAlgorithm, choices=kSearchAlgorithms)
        parent_parser.add_argument('-m', '--metric', metavar='',
                            help='Evaluation metric to be optimized',
                            required=True)
        parent_parser.add_argument('-e', '--extra_metrics', metavar='', nargs='+',
                            help='Additional metrics to be saved during the training',
                            required=False)
        parent_parser.add_argument('-o', '--mode', metavar='', 
                            help='Mode of optimization (either "min" or "max")', 
                            choices=kMetricMode,
                            required=True)
        parent_parser.add_argument('-s','--scheduler', metavar='',
                            help='Trial scheduling method for hyperparameter optimization. '+\
                                 'Allowed values: {}'.format(','.join(kSchedulers)),
                            default=kDefaultScheduler, choices=kSchedulers)
        parent_parser.add_argument('-n','--num_trials', metavar='',
                            help='Number of trials (search points)', 
                            type=int, required=True)
        parent_parser.add_argument('-c','--max_concurrent', metavar='',
                            help='Maximum number of trials to be run concurrently', 
                            type=int, default=kDefaultMaxConcurrent)            
        parent_parser.add_argument('-l', '--log_dir', metavar='',
                            help='Logging directory',
                            default=kDefaultLogDir)
        parent_parser.add_argument('-v','--verbose', metavar='', type=int, 
                            default=kDefaultVerbosity,
                            help='Verbosity level of Ray Tune')
        parent_parser.add_argument('--stop', metavar='',
                            help='Stopping criteria for the training',
                            default=kDefaultStopping)
        parent_parser.add_argument('-r','--resource', metavar='',
                            help='Resource allocated to each trial',
                            default=kDefaultResource)            
        parent_parser.add_argument('--scheduler_param', metavar='',
                            help='Extra parameters given to the trial scheduler', 
                            default=kDefaultSchedulerParam)
        parent_parser.add_argument('--algorithm_param', metavar='',
                            help='Extra parameters given to the hyperparameter optimization algorithm',
                            default=kDefaultAlgorithmParam)
        parser = ConfigurationBase.get_parser_test('hpo_config',
            parent_parser=parent_parser,
            description='Manage configuration for hyperparameter optimization',
            usage='hpogrid hpo_config <action> <config_name> [<options>]')
        return parser
            
    
    def get_parser(self, action=None):
        parser = self.get_base_parser()        
        if action in kConfigAction:
            parser.add_argument('name', help= "Name given to the configuration file")
            parser.add_argument('-a','--algorithm', 
                                help='Algorithm for hyperparameter optimization', 
                                default=kDefaultSearchAlgorithm, choices=kSearchAlgorithms)
            parser.add_argument('-m', '--metric', metavar='',
                                help='Evaluation metric to be optimized',
                                required=True)
            parser.add_argument('-e', '--extra_metrics', metavar='', nargs='+',
                                help='Additional metrics to be saved during the training',
                                required=False)
            parser.add_argument('-o', '--mode', 
                                help='Mode of optimization (either "min" or "max")', 
                                choices=kMetricMode,
                                required=True)
            parser.add_argument('-s','--scheduler', 
                                help='Trial scheduling method for hyperparameter optimization',
                                default=kDefaultScheduler, choices=kSchedulers)
            parser.add_argument('-n','--num_trials', metavar='',
                                help='Number of trials (search points)', 
                                type=int, required=True)
            parser.add_argument('-c','--max_concurrent', metavar='',
                                help='Maximum number of trials to be run concurrently', 
                                type=int, default=kDefaultMaxConcurrent)            
            parser.add_argument('-l', '--log_dir', metavar='',
                                help='Logging directory',
                                default=kDefaultLogDir)
            parser.add_argument('-v','--verbose', metavar='', type=int, 
                                default=kDefaultVerbosity,
                                help='Verbosity level of Ray Tune')
            parser.add_argument('--stop', metavar='',
                                help='Stopping criteria for the training',
                                default=kDefaultStopping)
            parser.add_argument('-r','--resource', metavar='',
                                help='Resource allocated to each trial',
                                default=kDefaultResource)            
            parser.add_argument('--scheduler_param', metavar='',
                                help='Extra parameters given to the trial scheduler', 
                                default=kDefaultSchedulerParam)
            parser.add_argument('--algorithm_param', metavar='',
                                help='Extra parameters given to the hyperparameter optimization algorithm',
                                default=kDefaultAlgorithmParam)
        else:
            parser = super().get_parser(action)
        return parser


import sys, os
import argparse
import json

from hpogrid.components.defaults import *
from hpogrid.components import validation
from hpogrid import ConfigurationBase
from json import JSONDecodeError

class SearchSpaceConfiguration(ConfigurationBase):

    config_type = 'search_space'
    description = 'Manage configuration for hyperparameter search space'
    usage = 'hpogrid search_space <action> <config_name> <search_space_definition>'
    list_columns = ['Search Space Configuration']
    show_columns = ['Hyperparameters', 'Search Space']  
    json_interpreted = []
    
    @staticmethod
    def get_parser_test():
        parser, subparsers = ConfigurationBase.get_parser_test(
            description='Manage configuration for hyperparameter search space',
            usage='hpogrid search_space <action> <config_name> <search_space_definition>')
        for action in subparsers.choices:
            subparsers.choices[action].usage = subparsers.choices[action].usage.replace('<config_type>', 'search_space')
        for action in ['create', 'update', 'recreate']:
            parser_action = subparsers.choices.get(action, None)
            if parser_action is None:
                raise RuntimeError('Subparser {} not found'.format(action))
            parser_action.add_argument('name', metavar='',
                help='Name given to the search space file')  
            parser_action.add_argument('search_space',metavar='', 
                help='A json decodable string defining the search space')  
        return parser

    def get_parser(self, action=None):
        parser = self.get_base_parser()               
        if action in kConfigAction:     
            parser.add_argument('name', 
                help='Name given to the configuration file')  
            parser.add_argument('search_space', 
                help='A json decodable string defining the search space')  
        else:
            parser = super().get_parser(action)
        return parser

    def configure(self, args, action='create'):

        config = vars(args)
        config_name = config.pop('name', None)
        search_space = config.pop('search_space', None)

        try:
            search_space = json.loads(search_space)
        except JSONDecodeError:
            print('ERROR: Cannot to decode input string into json format. Please check your input.')
            return None

        if action == 'update':
            config_path = self.get_config_path(config_name)
            if not os.path.exists(config_path):
                raise FileNotFoundError('Search space file {} not found. Update aborted.'.format(config_path))
            old_serach_space = json.load(open(config_path))
            search_space = {**old_serach_space, **search_space}
        
        if validation.validate_search_space(search_space):
            self.save(search_space, config_name, action)
        return search_space


import copy
from typing import List, Union
from functools import partial

import numpy as np

import hyperopt as hpo
from hyperopt.fmin import generate_trials_to_calculate

from hpogrid.generator.base_generator import Generator
from hpogrid.search_space.hyperopt_space import HyperOptSpace


class HyperOptGenerator(Generator):
    def get_searcher(self, search_space, metric:str, mode:str, **args):
        searcher = hpo.tpe.suggest
        if 'gamma' in args:
            searcher = partial(searcher, gamma=gamma) 
        hyperopt_space = HyperOptSpace(search_space).get_search_space()
        self.domain = hpo.Domain(lambda spc: spc, hyperopt_space)
        self.trials = hpo.Trials()
        self.rstate = np.random.RandomState()
        return searcher

    def ask(self, n_points:int =None):
        points = []
        trial_ids = self.trials.new_trial_ids(n_points)
        self.trials.refresh()
        new_trials = self.searcher(trial_ids, self.domain, self.trials,
            self.rstate.randint(2**31 - 1))
        for trial in new_trials:
            config = hpo.base.spec_from_misc(trial["misc"])
            memo = self.domain.memo_from_config(config)
            point = hpo.pyll.rec_eval(self.domain.expr, memo=memo)
            points.append(copy.deepcopy(point))
        return points

    def tell(self, point, value):
        value = self._to_metric_values(value)
        fake_trial = generate_trials_to_calculate([point])
        fake_trial.refresh()
        trial = fake_trial._trials[0]
        trial['state'] = hpo.base.JOB_STATE_DONE
        trial['result'] = {"loss": self.signature * value, "status": "ok"}
        self.trials.insert_trial_doc(trial)
        self.trials.refresh()

    def clear_history(self):
        self.trials.delete_all()


from hpogrid._version import __version__
from hpogrid.components.grid_handler import GridHandler
from hpogrid.components.job_builder import JobBuilder
from hpogrid.utils.grid_site_info import GridSiteInfo
from hpogrid.idds_interface.steering import SteeringIDDS
from hpogrid.configuration.configuration_base import ConfigurationBase
from hpogrid.configuration.hpo_configuration import HPOConfiguration
from hpogrid.configuration.grid_configuration import GridConfiguration
from hpogrid.configuration.search_space_configuration import SearchSpaceConfiguration
from hpogrid.configuration.project_configuration import ProjectConfiguration
from hpogrid.configuration.model_configuration import ModelConfiguration
#from hpogrid.components.panda_task_manager import PandaTaskManager
from hpogrid.utils.helper import get_workdir, load_configuration
from hpogrid.components.environment_settings import *
from hpogrid.components.core import *


import skopt
from skopt.space import Real, Categorical, Integer

from hpogrid.search_space.base_space import BaseSpace

class SkOptSpace(BaseSpace):

    def __init__(self, search_space = None):
        self.library = 'skopt'
        super().__init__(search_space)    
        

    def reset_space(self):
        self.search_space = []

    def append(self, space_value):
        self.search_space.append(space_value)

    def categorical(self, label, categories, grid_search = False):
        if grid_search != False:
            raise ValueError('{} does not allow grid search'.format(self.library))
        return Categorical(name=label, categories=categories)

    def uniformint(self, label, low, high):
        return Integer(name=label, low=low, high=high, prior="uniform")

    def uniform(self, label, low, high):
        return Real(name=label, low=low, high=high, prior='uniform')

    def loguniform(self, label, low, high, base=10):
        return Real(name=label, low=low, high=high, prior='log-uniform', base=base)

    def loguniformint(self, label, low, high, base=10):
        return Integer(name=label, low=low, high=high, prior="log-uniform", base=base)

    def fixed(self, label, value):
        return self.categorical(label=label, categories=[value])
    
    @staticmethod
    def transform(search_points, reference=None):
        if not isinstance(search_points, list):
            search_points = [search_points]
        sp_hyperparams = list(search_points[0])
        if not all(set(sp_hyperparams)==set(sp) for sp in search_points):
            raise ValueError('Inconsistent list of hyperparamters among'
                             ' search points')
        if reference and (set(reference) != set(sp_hyperparams)):
            raise ValueError('Inconsistent hyperparamters between search points'
                             ' and search space')
        # always respect the order of reference search space
        hyperparameters = list(reference) if reference else sp_hyperparams
        initial_points = []
        for sp in search_points:
            initial_points.append([sp[hp] for hp in hyperparameters])
        return initial_points


from past.utils import old_div
import numpy as np

from ray import tune

from hpogrid.search_space.base_space import BaseSpace

class TuneSpace(BaseSpace):

    def __init__(self, search_space = None):
        self.library = 'tune'
        super().__init__(search_space)    
        

    def reset_space(self):
        self.search_space = {}

    def append(self, space_value):
        self.search_space[space_value[0]] = space_value[1]

    def categorical(self, label, categories, grid_search = False):
        method = None
        if grid_search:
            method = tune.grid_search(categories)
        else:
            method = tune.choice(categories)
        return (label, method)

    def uniform(self, label, low, high):
        return (label, tune.uniform(low, high))
    
    def uniformint(self, label, low, high):
        return (label, tune.randint(low, high))

    def loguniform(self, label, low, high, base=10):
        return (label, tune.loguniform(low, high, base))

    def qloguniform(self, label, low, high, base=10):
        logmin = np.log(low) / np.log(base)
        logmax = np.log(high) / np.log(base)
        return (label, tune.sample_fram(lambda _: 
            np.round(old_div(base**(np.random.uniform(logmin, logmax)),q))*q))

    def quniform(self, label, low, high, q):
        return (label, tune.sample_from(lambda _: 
            np.round(old_div(np.random.uniform(low, high),q))*q))

    def quniformint(self, label, low, high, q):
        return (label, tune.sample_from(lambda _: 
            np.round(old_div(np.random.uniform(low, high),q))*q))

    def normal(self, mu, sigma):
        return (label, tune.sample_from(lambda _: sigma*np.random.randn()+mu))

    def qnormal(self, mu, sigma, q):
        return (label, tune.sample_from(lambda _: 
            np.round(old_div(sigma*np.random.randn()+mu,q))*q))

    def lognormal(self, mu, sigma, base = np.e):
        return (label, tune.sample_from(lambda _: 
            np.power(sigma*np.random.randn()+mu, base))) 

    def qlognormal(self, mu, sigma, q, base = np.e):
        return (label, tune.sample_from(lambda _: 
            np.round(old_div(np.power(sigma*np.random.randn()+mu, base),q))*q ))         

    def fixed(self, value):
        return (label, value)
                


import copy
from typing import List, Dict

import nevergrad as ng

from hpogrid.generator.base_generator import Generator
from hpogrid.search_space.nevergrad_space import NeverGradSpace

default_budget = 100
default_method = 'RandomSearch'

class NeverGradGenerator(Generator):
    def get_searcher(self, search_space:Dict, metric:str, mode:str, **args):
        search_space = NeverGradSpace(search_space).get_search_space()
        if ('method' in args) and (args['method']):
            method = args['method']
        else:
            method = default_method
        searcher = ng.optimizers.registry[method](
                parametrization=search_space, budget=default_budget)
        return searcher

    def ask(self, n_points:int = None):
        points = []
        for _ in range(n_points):
            point = self.searcher.ask().kwargs
            points.append(copy.deepcopy(point))
        return points

    def tell(self, point:Dict, value):
        value = self._to_metric_values(value)
        self.searcher.suggest(**point)
        candidate = self.searcher.ask()
        self.searcher.tell(candidate, self.signature * value)



import numpy as np 
import json
from pdb import set_trace

class BaseSpace():

    test_space = {
        'activation' : {
            'method': 'categorical',
            'dimension': {
                'categories': ['relu', 'sigmoid'],
                'grid_search' : 0
            }
        },
        'learning_rate' : {
            'method': 'loguniform',
            'dimension': {
                'low': 1e-6,
                'high': 1e-1,
                'base': 10
            }
        },
        'batchsize': {
            'method': 'categorical',
            'dimension': {
                'categories': [64,128,256,512,1024,2048],
                'grid_search' : 0
            }
        },
        'max_depth': {
            'method': 'uniformint',
            'dimension': {
                'low': 1,
                'high': 30
            }
        },
        'subsample': {
            'method': 'uniform',
            'dimension': {
                'low': 0.1,
                'high': 1.0
            }
        }
    }

    test_space_base_e = {
        'activation' : {
            'method': 'categorical',
            'dimension': {
                'categories': ['relu', 'sigmoid'],
                'grid_search' : 0
            }
        },
        'learning_rate' : {
            'method': 'loguniform',
            'dimension': {
                'low': np.e**-11,
                'high': np.e**-6,
                'base': np.e
            }
        },
        'batchsize': {
            'method': 'categorical',
            'dimension': {
                'categories': [64,128,256,512,1024,2048],
                'grid_search' : 0
            }
        },
        'max_depth': {
            'method': 'uniformint',
            'dimension': {
                'low': 1,
                'high': 30
            }
        },
        'subsample': {
            'method': 'uniform',
            'dimension': {
                'low': 0.1,
                'high': 1.0
            }
        }
    }    

    def __init__(self, search_space = None):
        self.reset_space()
        if search_space is not None:
            self.create(search_space)


    def get_search_space(self):
        return self.search_space
    
    def reset_space(self):
        self.library = 'base'
        self.search_space = {}

    def _method_error(self, method):
        raise ValueError('the method {} is not supported in {}'.format(method, self.library))

    def import_json(file):
        self.search_space = json.load(open(file))        
        
    def create(self,search_space = None):
        if search_space is None:
            search_space = self.search_space
        elif isinstance(search_space, str):
            search_space = json.loads(search_space)

        methods_map = {
            'categorical': self.categorical,
            'uniform': self.uniform,
            'uniformint': self.uniformint,
            'quniform': self.quniform,
            'loguniform': self.loguniform,
            'qloguniform': self.qloguniform,
            'normal': self.normal,
            'qnormal': self.qnormal,
            'lognormal': self.lognormal,
            'qlognormal': self.qlognormal,
            'fixed': self.fixed,
        }

        for hp in search_space:
            method = search_space[hp]['method']
            if method not in methods_map:
                self._method_error(method)
            space_value = methods_map[method](label = hp, **search_space[hp]['dimension'])
            if space_value is None:
                self._method_error(method)
            self.append(space_value)
        return self.search_space


    def append(self, space_value):
        self.search_space[space_value[0]] = space_value[1]

    def categorical(self, label, categories, grid_search = False):
        '''
        returns one of the categories, which should be a list
        '''
        space_value = (label, {'method': 'categorical', 
                               'dimension': {
                                   'categories': categories,
                                   'grid_search': grid_search}})
        return space_value

    def uniformint(self, label, low, high):
        '''
        returns a random integer in the range [low, high).
        '''
        space_value = (label, {'method': 'uniformint', 
            'dimension': { 'low': low, 'high': high}})
        return space_value

    def uniform(self, label, low, high):
        '''
        returns a value uniformly distributed between low and high.
        '''
        space_value = (label, {'method': 'uniform', 
            'dimension': { 'low': low, 'high': high}})
        return space_value

    def quniform(self, label, low, high, q):
        '''
        returns a value like round(uniform(low, high) / q) * q
        '''
        space_value = (label, {'method': 'quniform', 
            'dimension': { 'low': low, 'high': high, 'q': q}})
        return space_value

    def loguniform(self, label, low, high, base):
        '''
        loguniform returns a value drawn according to power(uniform(low, high), base = base) 
        so that the logarithm of the return value is uniformly distributed
        '''
        space_value = (label, {'method': 'loguniform', 
            'dimension': { 'low': low, 'high': high, 'base': base}})
        return space_value

    def qloguniform(self, label, low, high, q, base):
        '''
        Returns a value like round(exp(uniform(low, high)) / q) * q
        '''
        space_value = (label, {'method': 'qloguniform', 
            'dimension': { 'low': low, 'high': high, 'q': q, 'base': base}})
        return space_value

    def normal(self, label, mu, sigma):
        '''
        returns a real value that's normally-distributed with mean mu and standard deviation sigma
        '''
        space_value = (label, {'method': 'normal', 
            'dimension': { 'mu': mu, 'sigma': sigma}})
        return space_value


    def lognormal(self, label, mu, sigma, base):
        '''
        returns a value drawn according to power(normal(mu, sigma), base) so that the logarithm of the 
        return value is normally distributed
        '''
        space_value = (label, {'method': 'lognormal', 
            'dimension': { 'mu': mu, 'sigma': sigma, 'base': base}})
        return space_value

    def qnormal(self, label, mu, sigma, q):
        '''
        returns a value like round(normal(mu, sigma) / q) * q
        '''
        space_value = (label, {'method': 'qnormal', 
            'dimension': { 'mu': mu, 'sigma': sigma, 'q': q}})
        return space_value

    def qlognormal(self, label, mu, sigma, q, base):
        '''
        returns a value like round(power(normal(mu, sigma), base) / q) * q
        '''
        space_value = (label, {'method': 'qlognormal', 
            'dimension': { 'mu': mu, 'sigma': sigma, 'q': q, 'base': base}})
        return space_value

    def fixed(self, value):
        '''
        returns a fixed value
        '''
        space_value = (label, {'method': 'fixed', 'dimension': value})
        return space_value
    
    @classmethod
    def transform(cls, search_points):
        if not isinstance(search_points, list):
            search_points = [search_points]
        import pandas as pd
        search_space = pd.DataFrame(search_points).to_dict('list')
        base_space = cls()
        for hp in search_space:
            base_space.append(base_space.categorical(hp, search_space[hp]))
        return base_space.search_space


from typing import Dict, Optional, Union

from hpogrid.utils import helper

def submit(config_input:Union[Dict, str], n_jobs:int=1,
           site:Optional[str]=None, time:int=-1, mode:str='grid'):
    from hpogrid import GridHandler
    GridHandler.submit_job(config_input,
                           n_jobs=n_jobs,
                           site=site,
                           time=time)

def sites(**kwargs):
    from hpogrid import GridSiteInfo
    GridSiteInfo.show(**kwargs)
    
    
def create_project(config:Dict, action:str='create'):
    from hpogrid import ProjectConfiguration
    ProjectConfiguration.save(config, action=action)
    
def run(config_input: [Dict, str],
        search_points=None,
        mode='local'):
    from hpogrid import JobBuilder
    job = JobBuilder.from_input(config_input=config_input,
                                search_points=search_points,
                                mode=mode)
    job.run()


import sys, os
import argparse
import json
import fnmatch

from hpogrid.utils import stylus
from hpogrid.components.defaults import *
from hpogrid.utils.helper import load_configuration, get_base_path, get_config_dir
from json import JSONDecodeError

kActionList = ['create', 'recreate', 'update', 'list', 'show', 'remove']  

class ConfigurationBase():
    
    config_type = 'SUPPRESS'
    description = 'Manage configuration'
    usage = 'hpogrid <config_type> <action> <config_name> [<options>]'
    list_columns = []
    show_columns = []
    json_interpreted = []

    def __init__(self):
        parser = self.get_parser() 
        args = parser.parse_args(sys.argv[2:3])
        action = args.action
        parser = self.get_parser(action)
        args = parser.parse_args(sys.argv[3:])

        if action in kConfigAction:
            self.configure(args, action)
        elif hasattr(self, action):
            getattr(self, action)(**vars(args))
        else:
            print('Unrecognized action: {}'.format(action))
            parser.print_help()
            exit(1)            

    def get_base_parser(self):
        parser = argparse.ArgumentParser(
            description=self.description,
            usage=self.usage) 
        return parser
    
    @staticmethod
    def get_parser_test(config_type, parent_parser=None, **kwargs):
        parser = argparse.ArgumentParser(**kwargs) 
        subparsers = parser.add_subparsers(help='actions')
        def get_usage_str(action):
            return 'hpogrid {} {} [<options>]'.format(config_type, action)
        # action = list
        parser_list = subparsers.add_parser('list', 
        description='List all configuration files',
        usage=get_usage_str('list'),
        prog=get_usage_str('list'))
        parser_list.add_argument('-e', '--expression', metavar='',
                help='Filter out configuration files that matches the expression',
                required=False)
        
        # action = show
        parser_show = subparsers.add_parser('show', 
        description='Display the content of a configuration file',
        usage=get_usage_str('show'),
        prog=get_usage_str('show'))
        parser_show.add_argument('name', help='Name of the configuration file to show')
        
        # action = remove
        parser_remove = subparsers.add_parser('remove', 
        description='Remove a configuration file',
        usage=get_usage_str('remove'),
        prog=get_usage_str('remove'))
        parser_remove.add_argument('name', help='Name of the configuration file to remove')
        
        if parent_parser is None:
            parent_parser = argparse.ArgumentParser(add_help=False)
        
        # action = create
        parser_create = subparsers.add_parser('create', 
        description='Create a new configuration file',
        usage=get_usage_str('create'),
        prog=get_usage_str('create'),
        parents=[parent_parser])
        
        # action = update
        parser_update = subparsers.add_parser('update', 
        description='Update a configuration file',
        usage=get_usage_str('update'),
        prog=get_usage_str('update'),
        parents=[parent_parser])
        
        # action = recreate
        parser_recreate = subparsers.add_parser('recreate', 
        description='Recreate a configuration file',
        usage=get_usage_str('recreate'),
        prog=get_usage_str('recreate'),
        parents=[parent_parser])        
        return parser

    def get_parser(self, action=None):
        parser = self.get_base_parser()           
        if not action:
            parser.add_argument('action', help='Action to be performed', choices=kActionList)    
        elif action == 'list':
            parser.add_argument('--expr', metavar='',
                help='Filter out config files that matches the expression')
        elif action == 'show':
            parser.add_argument('name', help='Name of config file to show')
        elif action == 'remove':
            parser.add_argument('name', help='Name of config file to remove')
        else:
            raise ValueError('Unknown method: {}'.format(action))
        return parser

    def get_updated_config(self, config):
        non_updated_keys = []
        for key in config:
            if config[key] is None:
                non_updated_keys.append(key)
        for key in non_updated_keys:
            config.pop(key, None)
        config_path = self.get_config_path(config['name'])
        if not os.path.exists(config_path):
            raise FileNotFoundError('Configuration file {} not found. Update aborted.'.format(config_path))
        old_config = json.load(open(config_path))
        config = {**old_config, **config}
        return config

    def _retain_only_updated_options(self):
        parser = self.get_parser('update')
        for action in parser._actions:
            if (len(action.option_strings) > 0) and (action.default != '==SUPPRESS=='):
                action.default=None
        args = parser.parse_args(sys.argv[3:])
        return args

    def configure(self, args, action='create'):
        if action == 'update':
            args = self._retain_only_updated_options()

        config = vars(args)
        
        if action == 'update':
            config = self.get_updated_config(config)

        self.process_config(config)
        config_name = config.pop('name', None)
        if config is not None:
            self.save(config, config_name, action)
        return config

    def process_config(self, config):
        for key in config:
            if isinstance(config[key], bool):
                config[key] = int(config[key])            
            if (key in self.json_interpreted) and isinstance(config[key], str):
                try:
                    config[key] = json.loads(config[key])
                except JSONDecodeError:
                    raise ValueError('ERROR: Cannot decode the value of {} into json format.'
                        'Please check your input.'.format(key))   
        return config

    def save(self, config, name, action='create'):

        config_path = self.get_config_path(name)
        if (os.path.exists(config_path)) and (action=='create'):
            print('ERROR: {} configuration with name {} already exists.'
                'If you want to overwrite, use "recreate" or "update" action instead of "create".'.format(
                self.config_type, name))
        else:
            with open(config_path, 'w') as config_file:
                json.dump(config, config_file, indent=2)
            action_map = { 'create': 'Created', 'recreate': 'Recreated', 'update': 'Updated'}
            print('INFO: {} {} configuration {}'.format(action_map[action], self.config_type, config_path))
            self.show(name)

    @classmethod
    def get_config_dir(cls, force_create:bool=True):
        """Returns the directory under which the specific type of configuration files
        are stored
        
        Args:
            force_create: boolean, default=True
                Whether to force create the directory if not already exists
        """
        config_dir = get_config_dir(cls.config_type)

        if (not os.path.exists(config_dir)) and force_create:
            os.makedirs(config_dir, exist_ok=True)
        return config_dir
    
    @classmethod
    def get_config_path(cls, config_name:str=None, extension='.json'):
        """Returns the full path of a configuration file
        
        Args:
            config_name: str
                Name of the configuration file
            extension: str
                File extension for configuration file
        """

        config_dir = cls.get_config_dir()
        config_base_name = '{}{}'.format(config_name, extension)
        config_path = os.path.join(config_dir, config_base_name)
        return config_path
    
    @classmethod
    def get_config_list(cls, expr:str=None):
        """Returns a list of configuration files for a specific type of configuration
        
        Args:
            expr: str
                Regular expression for filtering name of configuration files
        """
        expr = expr if expr is not None else '*'
        
        config_dir = cls.get_config_dir()
        config_list = [os.path.splitext(d)[0] for d in os.listdir(config_dir) if not d.startswith('.')]
        if expr is not None:
            config_list = fnmatch.filter(config_list, expr)
        return config_list
    
    @classmethod
    def load(cls, name:str):
        """Returns a specified configuration file
        
        Args:
            name: str
                Name of configuration file
        """
        config_path = cls.get_config_path(name)
        config = load_configuration(config_path)
        return config

    @classmethod
    def list(cls, expr:str=None):
        """List out configuration files for a specific type of configuration as a table
        
        Args:
            expr: str
                Regular expression for filtering name of configuration files        
            exclude: list[str]
                Configuration files to exclude from listing
        """
        config_list = cls.get_config_list(expr)
        table = stylus.create_table(config_list, cls.list_columns)
        print(table)
    
    @classmethod
    def show(cls, name:str):
        """Display the content of a configuration file
        
        Args:
            name: str
                Name of configuration file        
        """
        config = cls.load(name)
        #table = stylus.create_table(config.items(), self.show_columns, indexed=False)
        table = stylus.create_formated_dict(config, cls.show_columns, indexed=False)
        print(table)  
        
    @classmethod
    def remove(cls, name):
        """Removes a configuration file
        
        Args:
            name: str
                Name of the configuration file to remove
        """
        config_path = cls.get_config_path(name)
        if os.path.exists(config_path):
            os.remove(config_path)
            print('INFO: Removed file {}'.format(config_path))
        else:
            print('ERROR: Cannot remove file {}. File does not exist.'.format(config_path))


import sys, os
import argparse
import json

from hpogrid.components.defaults import *
from hpogrid import ConfigurationBase

class ValidateSites(argparse.Action):
    def __call__(self, parser, args, values, option_string=None):
        if kGPUGridSiteList:
            for site in values:
                if site not in kGPUGridSiteList:
                    raise ValueError('Invalid grid site {}. '
                        'Please choose one of {}'.format(site, kGPUGridSiteList))
        else:
            print('INFO: Skipping validation of input grid site(s)'
                  ' in environment outside lxplus.')
            
        setattr(args, self.dest, ','.join(values))
                               
class GridConfiguration(ConfigurationBase):
    
    config_type = 'grid'
    description = 'Manage configuration for grid job submission'
    usage = 'hpogrid grid_config <action> <config_name> [<options>]'
    list_columns = ['Grid Configuration']
    show_columns = ['Attribute', 'Value']  
    json_interpreted = []
    
    @staticmethod
    def get_parser_test():
        parser, subparsers = ConfigurationBase.get_parser_test(
            description='Manage configuration for grid job submission',
            usage='hpogrid grid_config <action> <config_name> [<options>')
        for action in subparsers.choices:
            subparsers.choices[action].usage = subparsers.choices[action].usage.replace('<config_type>', 'grid_config')
        for action in ['create', 'update', 'recreate']:
            parser_action = subparsers.choices.get(action, None)
            if parser_action is None:
                raise RuntimeError('Subparser {} not found'.format(action))
            parser_action.add_argument('name', help = "Name given to the configuration file")
            parser_action.add_argument('-s', '--site', nargs='+',
                                help='Name of the grid site to where the jobs are submitted',
                                required=False, default=kDefaultGridSite,
                                action=ValidateSites)
            parser_action.add_argument('-c', '--container', metavar='',
                                help='Name of the docker or singularity container in which the jobs are run', 
                                required=False, default=kDefaultContainer)
            parser_action.add_argument('-i', '--inDS', metavar='',
                                help='Name of input dataset')
            parser_action.add_argument('-o', '--outDS', metavar='',
                                help='Name of output dataset', 
                                default=kDefaultOutDS)       
            parser_action.add_argument('-r', '--retry',
                                help='Check to enable retrying faild jobs',
                                action='store_true')
        return parser

    def get_parser(self, action=None):
        parser = self.get_base_parser()
        if action in kConfigAction:         
            parser.add_argument('name', help = "Name given to the configuration file")
            parser.add_argument('-s', '--site', nargs='+',
                                help='Name of the grid site to where the jobs are submitted',
                                required=False, default=kDefaultGridSite,
                                action=ValidateSites)
            parser.add_argument('-c', '--container', metavar='',
                                help='Name of the docker or singularity container in which the jobs are run', 
                                required=False, default=kDefaultContainer)
            parser.add_argument('-i', '--inDS', metavar='',
                                help='Name of input dataset')
            parser.add_argument('-o', '--outDS', metavar='',
                                help='Name of output dataset', 
                                default=kDefaultOutDS)       
            parser.add_argument('-r', '--retry',
                                help='Check to enable retrying faild jobs',
                                action='store_true')       
        else:
            parser = super().get_parser(action)
        return parser



import copy
from typing import List, Dict

from ax.service.ax_client import AxClient

from hpogrid.generator.base_generator import Generator
from hpogrid.search_space.ax_space import AxSpace

class GridGenerator(Generator):
    def get_searcher(self, search_space:Dict, metric:str, mode:str, **args):
        labels = list(search_space)
        # check that all samplings methods are categorical
        for label in labels:
            if search_space[label]['method'] != 'categorical':
                raise ValueError('Search space must be defined by categorical values for grid search')
        category_combinations = [search_space[label]['dimension']['categories'] for label in labels]
        
        search_points = list(set(itertools.product(*category_combinations)))
        self.labels = labels
        return search_points

    def tell(self, point:Dict, value):
        # search point values
        value = tuple([point[label] for label in self.labels])
        if value in self.searcher:
            self.searcher.remove(value)

    def ask(self, n_points:int =None):
        points = []
        if n_points > len(self.searcher):
            n_points = len(self.searcher)
        for i in range(n_points):
            points.append({ k:v for k,v in zip(self.labels, self.searcher[i])})
        return points


import json

kHPOGridEnvPath = 'HPOGRID_BASE_PATH'

kProjectConfigNameJson = 'project_config.json'
kProjectConfigNameYaml = 'project_config.yaml'

kiDDSConfigName = 'idds_config.json'
kiDDSSearchSpaceName = 'search_space.json'
kiDDSHPinput = 'input.json'
kiDDSHPoutput = 'output.json'
kiDDSMetrics = "metrics.tgz"

kConfigAction = ['create', 'recreate', 'update']

kAlgoMap = {
    'random': 'tune',
    'bayesian': 'hyperopt'
}

kDefaultiDDSConfigPath = '/afs/cern.ch/work/c/chlcheng/public/hpogrid/idds.cfg'
kDefaultProxyPath = '/etc/pki/tls/cert.pem'
#####################################################
##############    Paths Related      ################
#####################################################
kConfigPaths = {
    'project': 'projects',
    'hpo': 'config/hpo',
    'search_space': 'config/search_space',
    'model': 'config/model',
    'grid': 'config/grid'
}

kDockerBasePath = '/workDir'
kiDDSBasePath = kDockerBasePath
kGridSiteWorkDir = '/srv/workDir/'

#####################################################
############    Grid Site Related      ##############
#####################################################

try:
    from hpogrid.utils.grid_site_info import GridSiteInfo
    kGPUGridSiteList = GridSiteInfo.list_sites()
    kGPUGridSiteList += ['ANALY_CERN-PTEST','ANY', 'UKI-NORTHGRID-MAN-HEP']
except:
    kGPUGridSiteList = None

kStableGPUGridSiteList = ['ANALY_MANC_GPU_TEST', 'ANALY_MWT2_GPU', 'DESY-HH_GPU']
kDefaultGridSite = ','.join(kStableGPUGridSiteList)

kGPUSiteNGPU = {
    'ANALY_MANC_GPU_TEST': 10, #single queue, no submission parameters, 1 GPU per job
    'ANALY_QMUL_GPU_TEST': 6, # GPUNumber=x for now is hardcoded in the dev APF JDL,number of GPUs per job limited by cgroups, K80=2*K40, so total of 6 gpu slots avalable.
    'ANALY_MWT2_GPU': 8, #single queue, no submission parameters, 1 GPU per job
    'ANALY_BNL_GPU_ARC': 12, #also shared with Jupyter users who have priority
    'ANALY_INFN-T1_GPU': 2 #single queue, no submission parameters, 1 GPU per job
}

#########################################################
#############    Configuration Defaults    ##############
#########################################################

kDefaultContainer = '/cvmfs/unpacked.cern.ch/gitlab-registry.cern.ch/aml/hyperparameter-optimization/alkaid-qt/hpogrid:latest'
kDefaultContainer2 = '/cvmfs/unpacked.cern.ch/gitlab-registry.cern.ch/clcheng/hyperparameter-optimization-on-the-grid:latest'
kDefaultiDDSContainer = 'gitlab-registry.cern.ch/aml/hyperparameter-optimization/alkaid-qt/hpogrid:latest'
kDefaultContainer4 = 'docker://gitlab-registry.cern.ch/aml/hyperparameter-optimization/alkaid-qt/hpogrid:base'

kSearchAlgorithms = ['hyperopt', 'skopt', 'bohb', 'ax', 'tune', 'random', 'grid', 'bayesian', 'nevergrad']
kSchedulers = ['asynchyperband', 'bohbhyperband', 'pbt']
kMetricMode = ['min', 'max']





kDefaultOutDS = 'user.${{RUCIO_ACCOUNT}}.hpogrid.{HPO_PROJECT_NAME}.out.$(date +%Y%m%d%H%M%S)'

kHPOGridMetadataFormat = ['task_time_s', 'start_timestamp', 'project_name', 
                          'result', 'hyperparameters', 'metric', 'start_datetime', 
                          'best_config', 'end_datetime', 'mode']

kProjConfigFormat = {
    'project_name': {
        'required': True,
        'type': str
    },
    'scripts_path': {
        'required': True,
        'type': str
    },
    'model_config': {
        'required': True,
        'type': dict
    },
    'search_space': {
        'required': True,
        'type':dict
    },
    'hpo_config': {
        'required': True,
        'type':dict
    },
    'grid_config': {
        'required': True,
        'type':dict
    }
}

kConfigFormat = {
    'hpo':{
        'algorithm': {
            'required': False,
            'type': str,
            'choice': kSearchAlgorithms,
            'default': 'random'
        },
        'scheduler': {
            'required': False,
            'type': str,
            'choice': kSchedulers,
            'default': 'asynchyperband'
        },        
        'metric' : {
            'required': True,
            'type': str,
        },
        'extra_metrics': {
            'required': False,
            'type': (list, type(None)),
            'default': None
        },
        'mode' : {
            'required': True,
            'type': str,
            'choice': kMetricMode,
        },
        'resource': {
            'required': False,
            'type': (dict, type(None)),
            'default': None
        },
        'num_trials': {
            'required': True,
            'type': int,
        },
        'log_dir': {
            'required': False,
            'type': str,
            'default': './log'
        },
        'verbose': {
            'required': False,
            'type': int,
            'default': 0
        },
        'max_concurrent': {
            'required': False,
            'type': int,
            'default': 3
        },
        'stop': {
            'required': False,
            'type': dict,
            'default': {'training_iteration': 1}
        },
        'scheduler_param': {
            'required': False,
            'type': dict,
            'default': {}
        },
        'algorithm_param': {
            'required': False,
            'type': dict,
            'default': {}
        }       
    },
    'grid': {
        'site': {
            'required': False,
            'type': (list, str),
            'default': kDefaultGridSite
        },
        'container': {
            'required': False,
            'type': str,
            'default': kDefaultContainer
        },
        'retry': {
            'required': False,
            'type': int,
            'default': 0
        },
        'inDS': {
            'required': False,
            'type': (str, type(None)),
            'default': None
        },
        'outDS': {
            'required': False,
            'type': str,
            'default': kDefaultOutDS
        },
        'extra': {
            'required': False,
            'type': dict,
            'default': {}
        }
    },
    'model':{
        'script': {
            'required': True,
            'type': str
        },
        'model': {
            'required': True,
            'type': str
        },
        'param': {
            'required': False,
            'type': dict,
            'default': {}
        }
    },
    'search_space':{
        'categorical': {
            'categories' : { 'type': list, 'required': True},
            'grid_search' : { 'type': int, 'required': False, 'default': 0}
        },
        'uniform': {
            'low': { 'type': (float, int), 'required': True},
            'high': { 'type': (float, int), 'required': True}
        },
        'uniformint': {
            'low': { 'type': int, 'required': True},
            'high': { 'type': int, 'required': True}
        },
        'quniform': {
            'low': { 'type': (float, int), 'required': True},
            'high': { 'type': (float, int), 'required': True},
            'q': { 'type': (float, int), 'required': True}
        },
        'loguniform': {
            'low': { 'type': (float, int), 'required': True},
            'high': { 'type': (float, int), 'required': True},
            'base': { 'type': (float, int), 'required': False}
        },
        'qloguniform': {
            'low': { 'type': (float, int), 'required': True},
            'high': { 'type': (float, int), 'required': True},
            'q': { 'type': (float, int), 'required': True},
            'base': { 'type': (float, int), 'required': False}
        },
        'normal': {
            'mu': { 'type': (float, int), 'required': True},
            'sigma': { 'type': (float, int), 'required': True}
        },
        'qnormal': {
            'mu': { 'type': (float, int), 'required': True},
            'sigma': { 'type': (float, int), 'required': True},
            'q': { 'type': (float, int), 'required': True}
        },
        'lognormal': {
            'mu': { 'type': (float, int), 'required': True},
            'sigma': { 'type': (float, int), 'required': True},
            'base': { 'type': (float, int), 'required': False}
        },
        'qlognormal': {
            'mu': { 'type': (float, int), 'required': True},
            'sigma': { 'type': (float, int), 'required': True},
            'base': { 'type': (float, int), 'required': False},
            'q': { 'type': (float, int), 'required': True}
        },
        'fixed': {
            'value': { 'type': None, 'required': True}
        }    
    }
}

kDefaultGridExtraParam = json.dumps(kConfigFormat['grid']['extra']['default'])
kDefaultSearchAlgorithm = kConfigFormat['hpo']['algorithm']['default']
kDefaultScheduler = kConfigFormat['hpo']['scheduler']['default']
kDefaultMaxConcurrent = kConfigFormat['hpo']['max_concurrent']['default']
kDefaultLogDir = kConfigFormat['hpo']['log_dir']['default']
kDefaultVerbosity = kConfigFormat['hpo']['verbose']['default']
kDefaultStopping = json.dumps(kConfigFormat['hpo']['stop']['default'])
kDefaultResource = json.dumps(kConfigFormat['hpo']['resource']['default'])
kDefaultSchedulerParam = json.dumps(kConfigFormat['hpo']['scheduler_param']['default'])
kDefaultAlgorithmParam = json.dumps(kConfigFormat['hpo']['algorithm_param']['default'])
kDefaultModelParam = json.dumps(kConfigFormat['model']['param']['default'])

kHPOGridMetadataFormat = ['project_name', 'metric', 'mode', 'task_time_s', 'result', 'hyperparameters', 'best_config', 'start_datetime', 'end_datetime', 'start_timestamp']

kGridSiteMetadataFileName = 'userJobMetadata.json' #jobReport.json


import os
from hpogrid.components.defaults import *  

def get_run_mode():
    return os.environ.get('HPOGRID_RUN_MODE', 'LOCAL')

def get_datadir():
    return os.environ.get('HPOGRID_DATA_DIR', os.getcwd())

def get_workdir():
    return os.environ.get('HPOGRID_WORK_DIR', os.getcwd())

def grid_workdir():
    if os.path.exists(kGridSiteWorkDir):
        return kGridSiteWorkDir
    else:
        return os.getcwd()

def is_grid_job():
    return (get_run_mode() in ['GRID', 'IDDS'])

def is_idds_job():
    return (get_run_mode() == 'IDDS')

def is_local_job():
    return (get_run_mode() == 'LOCAL')
    
def setup(run_mode:str='local'):
    allowed_modes = ['local', 'grid', 'idds']
    run_mode = run_mode.lower()
    if run_mode not in allowed_modes:
        raise ValueError('Unknown running mode: {}. '
                         'Allowed running modes are {}'.format(run_mode, allowed_modes))
    print('INFO: Setting up HPO task')

    if run_mode == 'local':
        os.environ['HPOGRID_RUN_MODE'] = 'LOCAL'
        os.environ['HPOGRID_DATA_DIR'] = os.getcwd()
        os.environ['HPOGRID_WORK_DIR'] = os.getcwd()
    elif run_mode == 'grid':
        os.environ['HPOGRID_RUN_MODE'] = 'GRID'
        os.environ['HPOGRID_DATA_DIR'] = grid_workdir()
        os.environ['HPOGRID_WORK_DIR'] = grid_workdir()
    elif run_mode == 'idds':
        os.environ['HPOGRID_RUN_MODE'] = 'IDDS'
        os.environ['HPOGRID_DATA_DIR'] = grid_workdir()
        os.environ['HPOGRID_WORK_DIR'] = grid_workdir()
        
    print('INFO: HPO task will run in {} environment'.format(run_mode))
    print('INFO: Work directory is set to "{}"'.format(get_workdir()))
    print('INFO: Data directory is set to "{}"'.format(get_datadir()))
    
def reset():
    os.environ.pop('HPOGRID_RUN_MODE', None)
    os.environ.pop('HPOGRID_DATA_DIR', None)
    os.environ.pop('HPOGRID_WORK_DIR', None)


import os
import sys
import json
import time
import importlib
import argparse
from typing import Optional, Dict, List
from datetime import datetime

import numpy as np

import ray
from ray import tune

import hpogrid
from hpogrid.components.defaults import *
from hpogrid.utils import helper
from hpogrid.idds_interface import idds_utils


class JobBuilder():
    def __init__(self, parser=False):
        self.reset()
    
    @staticmethod       
    def get_parser():
        parser = argparse.ArgumentParser(description='Run hyperparameter optimization '
                                        'from a project/configuration file')     
        parser.add_argument('config_input', help='Name of project/configuration file')  
        parser.add_argument('-p', '--search_points', help='A json file containing the manual'
                            ' search points to run', default=None)    
        parser.add_argument('-m', '--mode',  help='Platform for running hyperparameter'
                            ' optimization. Choose from "local", "grid", "idds"',
                            choices=['local', 'grid', 'idds'],
                            default='local')   
        return parser

    @property
    def algorithm(self) -> str:
        return self._algorithm
    
    @algorithm.setter
    def algorithm(self, name):
        if name and (name not in kSearchAlgorithms):
            raise ValueError('unrecognized search algorithm: {}'.format(name))
        self._algorithm = kAlgoMap.get(name, name)
    
    @property
    def scheduler(self) -> str:
        return self._scheduler
    
    @scheduler.setter
    def scheduler(self, name):
        if name and (name not in kSchedulers):
            raise ValueError('unrecognized trial scheduler: {}'.format(name))
        self._scheduler = name
        

    @property
    def mode(self) -> str:
        return self._mode
    
    @mode.setter
    def mode(self, value):
        if value not in kMetricMode:
            raise ValueError('metric mode must be either "min" or "max"')
        self._mode = value
        
    @property
    def resource_per_trial(self):
        return self._resource_per_trial
    
    @resource_per_trial.setter
    def resource_per_trial(self, value):
        resource = JobBuilder.get_resource_info()
        resource_per_trial = {}
        for device in ['cpu', 'gpu']:
            if resource[device] >= self.max_concurrent:
                resource_per_trial[device] = int(resource[device]/self.max_concurrent)  
            elif resource[device] > 0:
                resource_per_trial[device] = resource[device]/self.max_concurrent
            else:
                resource_per_trial[device] = 0
        max_concurrent = self.max_concurrent
        if isinstance(value, dict):
            for device in ['cpu', 'gpu']:
                if device in value and value[device]:
                    resource_per_trial[device] = min(value[device], resource_per_trial[device])
        print('INFO: Each trial will use {} GPU(s) resource'.format(resource_per_trial['gpu']))
        print('INFO: Each trial will use {} CPU(s) resource'.format(resource_per_trial['cpu']))
        self._resource_per_trial = resource_per_trial

        
    @property
    def df(self) -> 'pandas.DataFrame':
        return self._df
    
    @property
    def start_datetime(self) -> str:
        return self._start_datetime
    
    @property
    def end_datetime(self) -> str:
        return self._end_datetime
    
    @property
    def start_timestamp(self) -> float:
        return self._start_timestamp
    
    @property
    def total_time(self) -> float:
        return self._total_time
    
    @property
    def best_config(self) -> Dict:
        return self._best_config

    @property
    def hyperparameters(self):
        return self._hyperparameters
    
    @property
    def log_dir(self) -> str:
        return self._log_dir

    @log_dir.setter
    def log_dir(self, val):
        """
        self._log_dir = val
        if hpogrid.is_local_job() and (not os.path.isabs(val)):
            project_path = helper.get_project_path(self.project_name)
            if os.path.exists(project_path):
                self._log_dir = os.path.join(project_path, val)
        self._log_dir = helper.pretty_path(self._log_dir)
        """
        self._log_dir = val
            
        
            
    def reset(self):
        self.project_name = None
        # model configuration
        self.model_script = None
        self.model_name = None
        self.model_param = kDefaultModelParam  
        # search space
        self.search_space = {}
        # hpo configuration
        self._algorithm = None
        self.metric = None
        self._mode = None
        self._scheduler = None  
        self.algorithm_param = kDefaultAlgorithmParam        
        self.scheduler_param = kDefaultSchedulerParam
        self.num_trials = None
        self.max_concurrent = kDefaultMaxConcurrent
        self.verbose = False
        self._log_dir = kDefaultLogDir
        self.stop = kDefaultStopping
        self._resource = None
        self.extra_metrics = None
        # parameters for logging
        self._df = None
        self._start_datetime = None
        self._end_datetime = None
        self._total_time = None
        self._best_config = None
        self._hyperparameters = {}
        self._start_timestamp = None
        self.idds_job = False
        self.scripts_path = os.getcwd()
        self.search_points = None
        self.summary_fname = kGridSiteMetadataFileName

        
    @staticmethod
    def get_scheduler(name, metric, mode, search_space = None, **args):
        if (name == None) or (name == 'None'):
            return None
        elif name == 'asynchyperband':
            from hpogrid.scheduler.asynchyperband_scheduler import AsyncHyperBandSchedulerWrapper
            return AsyncHyperBandSchedulerWrapper().create(metric, mode, **args)
        elif name == 'bohbhyperband':
            from hpogrid.scheduler.bohbhyperband_scheduler import BOHBHyperBandSchedulerWrapper
            return BOHBHyperBandSchedulerWrapper().create(metric, mode, **args)
        elif name == 'pbt':
            from hpogrid.scheduler.pbt_scheduler import PBTSchedulerWrapper
            if search_space is None:
                raise ValueError('Missing search space definition for pbt scheduler')
            return PBTSchedulerWrapper().create(metric, mode, search_space, **args)

    @staticmethod
    def get_search_space(base_search_space, algorithm):
        if base_search_space is None:
            raise ValueError('search space can not be empty')    
        if algorithm == 'ax':
            from hpogrid.search_space.ax_space import AxSpace
            return AxSpace(base_search_space).get_search_space()
        elif algorithm == 'bohb':
            from hpogrid.search_space.bohb_space import BOHBSpace
            return BOHBSpace(base_search_space).get_search_space()
        elif algorithm == 'hyperopt':
            from hpogrid.search_space.hyperopt_space import HyperOptSpace
            return HyperOptSpace(base_search_space).get_search_space()
        elif algorithm == 'skopt':
            from hpogrid.search_space.skopt_space import SkOptSpace
            return SkOptSpace(base_search_space).get_search_space()
        elif algorithm == 'tune':
            from hpogrid.search_space.tune_space import TuneSpace
            return TuneSpace(base_search_space).get_search_space()
        elif algorithm == 'nevergrad':
            from hpogrid.search_space.nevergrad_space import NeverGradSpace
            return NeverGradSpace(base_search_space).get_search_space() 
        else:
            raise ValueError('Unrecognized search algorithm: {}'.format(name))

    @staticmethod
    def get_algorithm(name, metric, mode, base_search_space, max_concurrent=None, **args):
        if name == 'ax':
            from hpogrid.algorithm.ax_algorithm import AxAlgoWrapper
            algorithm = AxAlgoWrapper().create(metric, mode, base_search_space, **args)
        elif name == 'bohb':
            from hpogrid.algorithm.bohb_algorithm import BOHBAlgoWrapper
            algorithm = BOHBAlgoWrapper().create(metric, mode, base_search_space, **args)
        elif name == 'hyperopt':
            from hpogrid.algorithm.hyperopt_algorithm import HyperOptAlgoWrapper
            algorithm = HyperOptAlgoWrapper().create(metric, mode, base_search_space, **args)
        elif name == 'skopt':
            from hpogrid.algorithm.skopt_algorithm import SkOptAlgoWrapper
            algorithm = SkOptAlgoWrapper().create(metric, mode, base_search_space, **args)
        elif name == 'nevergrad':
            from hpogrid.algorithm.nevergrad_algorithm import NeverGradAlgoWrapper
            algorithm = NeverGradAlgoWrapper().create(metric, mode, base_search_space, **args)
        elif name == 'tune':
            algorithm = None
        else:
            raise ValueError('Unrecognized search algorithm: {}'.format(name))
        # limit max concurrency
        if algorithm and max_concurrent:
            from ray.tune.suggest import ConcurrencyLimiter
            algorithm = ConcurrencyLimiter(algorithm, max_concurrent=max_concurrent) 
        return algorithm

    @staticmethod
    def get_model(script_name, model_name, scripts_path=None):
        scripts_path = os.getcwd() if scripts_path is None else scripts_path
        model = None
        script_name_noext = os.path.splitext(script_name)[0]
        try: 
            helper.set_scripts_path(scripts_path)
            module = importlib.import_module(script_name_noext)
            model = getattr(module, model_name)
        except: 
            raise ImportError('Unable to import function/class {} '
                'from training script: {}.py'.format(model_name, script_name_noext))
        finally:
            helper.set_scripts_path(scripts_path, undo=True)
        print('INFO: Loaded module {}'.format(model.__name__))
        return model

    def create_metadata(self, df) -> Dict:

        summary = {
            'project_name' : self.project_name,
            'start_datetime': self.start_datetime,
            'end_datetime': self.end_datetime,
            'start_timestamp': self.start_timestamp,
            'task_time_s' : self.total_time,
            'hyperparameters': self.hyperparameters,
            'metric': self.metric,
            'mode' : self.mode,
            'best_config' : self.best_config, 
        }
        
        rename_cols = { 'config/{}'.format(hp): hp for hp in self.hyperparameters}
        rename_cols['time_total_s'] = 'time_s'
        
        df = df.rename(columns=rename_cols)
        cols_to_save = ['time_s'] + self.hyperparameters
        if 'metric' in summary:
            cols_to_save.append(summary['metric'])
        if self.extra_metrics:
            cols_to_save += self.extra_metrics
        df = df.filter(cols_to_save, axis=1).transpose()
        summary['result'] = df.to_dict()

        return summary

    @staticmethod
    def get_resource_info() -> Dict:
        resource = {}
        n_gpu = helper.get_n_gpu()
        n_cpu = helper.get_n_cpu()
        print('INFO: Number of GPUs detected: ',n_gpu)
        print('INFO: Number of CPUs detected: ',n_cpu)
        resource['gpu'] = n_gpu
        resource['cpu'] = n_cpu
        return resource
    
    def run(self) -> None:
            
        self._start_timestamp = start = time.time()
        self._start_datetime = datetime.now().strftime("%m/%d/%Y, %H:%M:%S")

        # get model parameters
        if self.algorithm == 'tune':
            tune_config_space = self.get_search_space(self.search_space, algorithm='tune')
        else:
            tune_config_space = {}

        tune_config_space.update(self.model_param)

        # get search algorithm
        algorithm = self.get_algorithm(
            self.algorithm,
            self.metric,
            self.mode,
            self.search_space,
            self.max_concurrent,
            **self.algorithm_param)
        
        # get trial scheduler
        scheduler = self.get_scheduler(
            self.scheduler,
            self.metric,
            self.mode,
            self.search_space,
            **self.scheduler_param)
        
        # get model
        model = self.get_model(self.model_script, self.model_name, self.scripts_path)
        
        loggers = None
        #loggers = self.get_loggers()
        
        # save something to prevent looping job
        with open(kGridSiteMetadataFileName,'w') as output:
            json.dump({}, output)
            
        # HPO starts
        try:
            
            # extract input dataset if it is a tar file
            # (for grid jobs only)
            extracted_files = []
            if hpogrid.is_grid_job():
                datadir = hpogrid.get_datadir()
                extracted_files = helper.extract_tarball(datadir, datadir)
                
            ray.init(include_dashboard=False)
#           lru_evict=True            
            analysis = tune.run(
                model,
                name=self.project_name,
                scheduler=scheduler,
                search_alg=algorithm,
                config=tune_config_space,
                num_samples=self.num_trials,
                resources_per_trial=self.resource_per_trial,
                verbose=self.verbose,
                local_dir=self.log_dir,
                loggers=loggers,
                stop=self.stop)
#                sync_to_driver=False,
#                global_checkpoint_period=np.inf)
        finally:
            ray.shutdown()
            # remove the extracted tar files
            helper.remove_files(extracted_files)
        
        end = time.time()
        self._total_time = float(end-start)
        self._end_datetime = datetime.now().strftime("%m/%d/%Y, %H:%M:%S")

        self._best_config = analysis.get_best_config(metric=self.metric, mode=self.mode)
        
        print("Best config: ", self.best_config)
        print("Time taken in seconds: ", self.total_time)
        
        print("INFO: Ray Tune log files are saved at {}".format(self.log_dir))

        df = analysis.dataframe(metric=self.metric, mode=self.mode)

        metadata = self.create_metadata(df)
        
        summary_output_path = helper.pretty_path(self.summary_fname)
        # save metadata
        with open(summary_output_path, 'w') as output:
            print('INFO: Job metadata is saved at {}'.format(summary_output_path))
            json.dump(metadata, output, cls=helper.NpEncoder)
        
        # save idds output
        if hpogrid.is_idds_job():
            idds_utils.save_idds_output_from_metadata(metadata)

        # reset environment
        hpogrid.reset()
    
    def get_loggers(self):
        from ray.tune.logger import NoopLogger, MLFLowLogger
        return [NoopLogger, MLFLowLogger]
    
    @classmethod
    def from_input(cls, config_input: [Dict, str],
                    search_points=None,
                    mode='local'):
        
        hpogrid.setup(mode)
        
        input_as_project = helper.is_project(config_input)
        config = helper.load_configuration(config_input)
        
        # if input is a project name, set scripts path to 
        # where the scripts are stored in the project directory
        # if input is a configuration file, set scripts path to
        # the scripts path stated in configuration file
        if input_as_project:
            scripts_path = helper.get_scripts_path(config_input)
        else:
            scripts_path = helper.pretty_dirname(config['scripts_path'])
        
        job = JobBuilder()
        job.project_name = config['project_name']
        job.scripts_path = scripts_path
        
        # Load search space
        job.search_space = config['search_space']
        
        # Load hpo configuration
        job.algorithm = config['hpo_config']['algorithm']
        job.metric = config['hpo_config']['metric']
        job.mode = config['hpo_config']['mode']
        job.scheduler = config['hpo_config']['scheduler']
        job.scheduler_param = config['hpo_config']['scheduler_param']
        job.algorithm_param = config['hpo_config']['algorithm_param']
        job.num_trials = config['hpo_config']['num_trials']
        job.max_concurrent = config['hpo_config']['max_concurrent']
        job.verbose = config['hpo_config']['verbose']
        job.log_dir = config['hpo_config']['log_dir']
        job.stop = config['hpo_config'].get('stop', None)
        job.resource_per_trial = config['hpo_config'].get('resource', None)
        job.extra_metrics = config['hpo_config'].get('extra_metrics', None)
        
        # configuration for idds
        if (not search_points) and hpogrid.is_idds_job():
            search_points = idds_utils.get_search_points()
        
        if search_points:
            if isinstance(search_points, str):
                with open(search_points, 'r') as f:
                    search_points = json.load(f)
            job.search_points = search_points
            from hpogrid.search_space.skopt_space import SkOptSpace
            skopt_search_points = SkOptSpace.transform(search_points, 
                                      reference=job.search_space)
            job.algorithm = 'skopt'
            job.scheduler = None
            job.scheduler_param = {}
            job.algorithm_param = {'points_to_evaluate': skopt_search_points}
            job.num_trials = len(skopt_search_points)

        # Load model configuration
        job.model_script = config['model_config']['script']
        job.model_name = config['model_config']['model']
        job.model_param = config['model_config']['param']

        job._hyperparameters = list(job.search_space.keys())
    
        return job
    
